################################################################################
# Copyright (c) 2014-2019, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Perform on-device percentile calculation of 2D arrays."""
# see scripts/maskedsumtest.py for an example

from typing import Tuple, Mapping, Callable, Optional, Any, cast
from typing_extensions import TypedDict

import numpy as np

from . import accel
from . import tune
from .abc import AbstractContext, AbstractCommandQueue


_TuningDict = TypedDict('_TuningDict', {'size': int})


class MaskedSumTemplate:
    """Kernel for calculating masked sums of a 2D array of data.

    Masked sums are calculated per column (along rows, independently per column).

    Parameters
    ----------
    context
        Context for which kernels will be compiled
    use_amplitudes
        If true, the amplitudes of the inputs rather than the inputs
        themselves will be summed.
    tuning
        Kernel tuning parameters; if omitted, will autotune. The possible
        parameters are

        - size: number of workitems per workgroup
    """

    autotune_version = 1

    def __init__(self, context: AbstractContext, use_amplitudes: bool = False,
                 tuning: Optional[_TuningDict] = None) -> None:
        self.context = context
        self.use_amplitudes = use_amplitudes
        if tuning is None:
            tuning = self.autotune(context, use_amplitudes)
        self.size = tuning['size']
        self.program = accel.build(context, "maskedsum.mako", {
            'size': self.size,
            'use_amplitudes': use_amplitudes
        })

    @classmethod
    @tune.autotuner(test={'size': 256})
    def autotune(cls, context: AbstractContext, use_amplitudes: bool) -> _TuningDict:
        queue = context.create_tuning_command_queue()
        columns = 5000
        in_shape = (4096, columns)
        rs = np.random.RandomState(seed=1)
        host_data = rs.uniform(size=(in_shape[0], in_shape[1], 2)).astype(np.float32)
        host_data = host_data.view(dtype=np.complex64)[..., 0]
        host_mask = np.ones((in_shape[0],)).astype(np.float32)

        def generate(size: int) -> Callable[[int], float]:
            fn = cls(context, use_amplitudes, {
                'size': size}).instantiate(queue, in_shape)
            inp = fn.slots['src'].allocate(fn.allocator)
            msk = fn.slots['mask'].allocate(fn.allocator)
            fn.slots['dest'].allocate(fn.allocator)
            inp.set(queue, host_data)
            msk.set(queue, host_mask)
            return tune.make_measure(queue, fn)

        return cast(_TuningDict, tune.autotune(generate, size=[32, 64, 128, 256, 512, 1024]))

    def instantiate(self, command_queue: AbstractCommandQueue, shape: Tuple[int, int],
                    allocator: Optional[accel.AbstractAllocator] = None) -> 'MaskedSum':
        return MaskedSum(self, command_queue, shape, allocator)


class MaskedSum(accel.Operation):
    """Concrete instance of :class:`MaskedSumTemplate`.

    .. rubric:: Slots

    **src**
        Input type complex64
        Shape is number of rows by number of columns, masked sum is calculated
        along the rows, independently per column.

    **mask**
        Input type float32
        Shape is (number of rows of input).

    **dest**
        Output type complex64
        Shape is (number of columns of input)
    """

    def __init__(self, template: MaskedSumTemplate, command_queue: AbstractCommandQueue,
                 shape: Tuple[int, int],
                 allocator: Optional[accel.AbstractAllocator] = None) -> None:
        super().__init__(command_queue, allocator)
        self.template = template
        self.kernel = template.program.get_kernel("maskedsum_float")
        self.shape = shape
        self.slots['src'] = accel.IOSlot(
            (shape[0], accel.Dimension(shape[1], template.size)),
            np.complex64)
        self.slots['mask'] = accel.IOSlot((shape[0],), np.float32)
        self.slots['dest'] = accel.IOSlot(
            (accel.Dimension(shape[1], template.size),),
            np.float32 if template.use_amplitudes else np.complex64)

    def _run(self) -> None:
        src = self.buffer('src')
        mask = self.buffer('mask')
        dest = self.buffer('dest')
        self.command_queue.enqueue_kernel(
            self.kernel,
            [
                src.buffer, mask.buffer, dest.buffer,
                np.int32(src.padded_shape[1]), np.int32(src.shape[0])
            ],
            global_size=(accel.roundup(src.shape[1], self.template.size),),
            local_size=(self.template.size, ))

    def parameters(self) -> Mapping[str, Any]:
        return {
            'shape': self.slots['src'].shape,       # type: ignore
            'use_amplitudes': self.template.use_amplitudes
        }



#!/usr/bin/env python

################################################################################
# Copyright (c) 2015-2020, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import time
import numpy as np
from katsdpsigproc import accel, maskedsum

context = accel.create_some_context(True)
queue = context.create_command_queue(profile=True)

data = np.random.randn(4000, 5000, 2).astype(np.float32).view(dtype=np.complex64)[..., 0]
mask = np.ones((4000,)).astype(np.float32)

template = maskedsum.MaskedSumTemplate(context)
msum = template.instantiate(queue, data.shape)
msum.ensure_all_bound()
msum.buffer('src').set(queue, data)
msum.buffer('mask').set(queue, mask)
start_event = queue.enqueue_marker()
msum()
end_event = queue.enqueue_marker()
out = msum.buffer('dest').get(queue)

t0 = time.time()
expected = np.sum(data * mask.reshape(data.shape[0], 1), axis=0).astype(np.complex64)
t1 = time.time()
print('gpu:', end_event.time_since(start_event), 'cpu:', t1 - t0)
np.testing.assert_equal(out.reshape(-1), expected)



################################################################################
# Copyright (c) 2014-2020, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Library to contain 2d RFI flagging routines and other RFI related functions."""

import multiprocessing

import concurrent.futures
import numpy as np
import numba

from . import MAD_NORMAL


def _as_min_dtype(value):
    """Convert a non-negative integer into a numpy scalar of the narrowest type will hold it.

    This is used because in some cases an array must be allocated of the
    same type later, and using the narrowest type saves memory in that array.
    """
    if value >= 0 and value < 2**8:
        dtype = np.uint8
    elif value >= 0 and value < 2**16:
        dtype = np.uint16
    elif value >= 0 and value < 2**32:
        dtype = np.uint32
    else:
        dtype = np.int64
    return np.array(value, dtype)


def _asbool(data):
    """Create a boolean array with the same values as `data`.

    The `data` must contain only 0's and 1's. If possible, a view is returned,
    otherwise a copy.
    """
    if data.dtype.itemsize == 1:
        return data.view(np.bool_)
    else:
        return data.astype(np.bool_)


@numba.extending.overload(_asbool, jit_options=dict(nogil=True))
def _overload_asbool(data):
    if (isinstance(data.dtype, numba.types.Boolean)
            or (isinstance(data.dtype, numba.types.Integer) and data.dtype.bitwidth == 8)):
        return lambda data: data.view(np.bool_)
    else:
        return lambda data: data.astype(np.bool_)


@numba.jit(nopython=True, nogil=True)
def _average_freq(in_data, in_flags, factor):
    """Do several preconditioning steps.

    1. Converts complex data to real.
    2. Flags data with non-finite values.
    3. Sets the value of flagged elements to zero.
    4. Does frequency averaging by a factor of `factor`.
    5. Transposes the data ordering so that baseline is the first,
       slowest-varying axis.

    Parameters
    ----------
    in_data : ndarray, real or complex
        Visibilities or their magnitudes, with shape (time, frequency, baseline).
    in_flags : ndarray, bool
        Flags corresponding to the visibilities. This can safely be a type
        other than bool, where non-zero values indicate flagged data.
    factor : int
        Amount by which to decimate in frequency. This must be a numpy 0-d
        array (so that the dtype can be extracted).
    """
    if in_data.shape != in_flags.shape:
        raise ValueError('shape mismatch')
    n_time, n_freq, n_bl = in_data.shape
    a_freq = (n_freq + factor - 1) // factor
    out_shape = (n_bl, n_time, a_freq)
    avg_data = np.zeros(out_shape, np.float32)
    avg_weight = np.zeros(out_shape, factor.dtype)
    # TODO: might need to do this through a temporary buffer to avoid cache
    # aliasing problems.
    for i in range(n_time):
        for j in range(n_freq):
            jout = j // factor
            for k in range(n_bl):
                data = np.abs(in_data[i, j, k])
                if not in_flags[i, j, k] and not np.isnan(data):
                    avg_data[k, i, jout] += data
                    avg_weight[k, i, jout] += 1
    for i in range(n_bl):
        for j in range(n_time):
            for k in range(a_freq):
                flag = avg_weight[i, j, k] == 0
                if flag:
                    avg_data[i, j, k] = 0   # Avoid divide by zero and a NaN
                else:
                    avg_data[i, j, k] /= avg_weight[i, j, k]
                # Replace weight with flag (in-place) to save memory
                avg_weight[i, j, k] = flag
    return avg_data, _asbool(avg_weight)


@numba.jit(nopython=True, nogil=True)
def _time_median(data, flags):
    """Independently for each channel, compute the median of the unflagged values.

    If all values for a channel are flagged, 0 is used instead, and the result
    is flagged.

    The time dimension is kept in the result as a length-1 dimension.

    Parameters
    ----------
    data : ndarray, real
        Visibilities, with shape (time, frequency)
    flags : ndarray, bool
        Flags corresponding to `data`

    Returns
    -------
    out_data : ndarray, real
        Median of `data` for each frequency
    out_flags : ndarray, bool
        Flags corresponding to `out_data`
    """
    n_time, n_freq = data.shape
    values = np.empty((n_freq, n_time), data.dtype)
    counts = np.zeros(n_freq, np.uint32)
    out_data = np.empty((1, n_freq), data.dtype)
    out_flags = np.zeros((1, n_freq), np.bool_)
    for t in range(n_time):
        for f in range(n_freq):
            if not flags[t, f]:
                values[f, counts[f]] = data[t, f]
                counts[f] += 1
    for f in range(n_freq):
        if counts[f] == 0:
            out_data[0, f] = 0     # No data
            out_flags[0, f] = True
        else:
            out_data[0, f] = np.median(values[f, :counts[f]])
    return out_data, out_flags


@numba.jit(nopython=True, nogil=True)
def _median_abs(data, flags):
    """Compute median of absolute values of non-flagged values in `data`."""
    values = np.empty(data.size, data.dtype)
    n = np.int64(0)
    for idx in np.ndindex(data.shape):
        if not flags[idx]:
            values[n] = np.abs(data[idx])
            n += 1
    if n == 0:
        return np.nan
    else:
        return np.median(values[:n])


@numba.jit(nopython=True, nogil=True)
def _median_abs_axis0(data, flags):
    """Compute median of absolute values of non-flagged values in `data`, along axis 0.

    The first dimension is kept in the output as a dimension of size 1 (to
    avoid issues with numba converting 0d arrays to scalars.
    """
    values = np.empty(data.shape[1:] + data.shape[:1], data.dtype)
    counts = np.zeros(data.shape[1:], np.uint32)
    out_data = np.empty((1,) + data.shape[1:], data.dtype)
    for i in range(data.shape[0]):
        for j in np.ndindex(data.shape[1:]):
            if not flags[(i,) + j]:
                values[j + (counts[j],)] = np.abs(data[(i,) + j])
                counts[j] += 1
    for j in np.ndindex(data.shape[1:]):
        if counts[j] == 0:
            out_data[(0,) + j] = np.nan
        else:
            out_data[(0,) + j] = np.median(values[j + (slice(0, counts[j]),)])
    return out_data


@numba.jit(nopython=True, nogil=True)
def _linearly_interpolate_nans1d(data):
    """Replace NaNs in `data` by linear interpolation in-place.

    Extrapolation is done by repeating the first/last valid element.  If all
    input data are NaNs, they are all replaced by zeros.

    Parameters
    ----------
    data : ndarray, real
        Data to interpolate, 1D. It is modified in-place.
    """
    n = data.size
    # Find first valid value
    p = 0
    while p < n and np.isnan(data[p]):
        p += 1
    if p == n:
        data[:] = 0
        return
    data[:p] = data[p]     # Extrapolate backwards
    p += 1
    while p < n:
        if np.isnan(data[p]):
            # Find next valid value
            q = p + 1
            while q < n and np.isnan(data[q]):
                q += 1
            if q == n:
                data[p:] = data[p - 1]   # Extrapolate forwards
            else:
                start = data[p - 1]
                grad = (data[q] - start) / (q - (p - 1))
                for i in range(p, q):
                    data[i] = start + (i - (p - 1)) * grad
            p = q
        else:
            p += 1


@numba.jit(nopython=True, nogil=True)
def _linearly_interpolate_nans(data):
    """Replace nans in `data` by linear interpolation across frequencies.

    Extrapolation is done by repeating the first/last valid element.

    Parameters
    ----------
    data : ndarray, real
        Data to interpolate, with shape (time, frequency).
    """
    for i in range(data.shape[0]):
        _linearly_interpolate_nans1d(data[i])


@numba.jit(nopython=True, nogil=True)
def _box_gaussian_filter1d(data, r, out, passes):
    """Implement :func:`_box_gaussian_filter` along the first axis of an array.

    It is safe to use this function in-place i.e. with `out` equal to `data`.

    Parameters
    ----------
    data : ndarray, real
        Input data, with at least 1 dimension.
    r : int
        Radius of the box filter
    out : ndarray, real
        Output data, with same shape as `data`.
    passes : int
        Number of boxcar filters to apply
    """
    K = passes
    if data.shape[0] == 0 or K == 0:
        out[:] = data[:]
        return
    d = 2 * r + 1
    # Pad on left with zeros.
    padding = r * K
    # TODO: hoist memory allocations into caller
    padded = np.empty((data.shape[0] + padding,) + data.shape[1:], data.dtype)
    padded[:padding] = 0
    padded[padding:] = data
    prev_start = padding   # First element with valid data
    s = np.zeros(data.shape[1:], np.float64)
    for p in range(1, K + 1):
        # On each pass, padded[i] is replaced by the sum of padded[i : i + d]
        # from the previous pass. The accumulator is kept in double precision
        # to avoid excessive accumulation of errors.
        s[()] = 0
        start = padding - 2 * r * p
        stop = start + data.shape[0] + 2 * padding
        start = max(start, 0)
        stop = min(stop, padded.shape[0])
        tail = min(stop, padded.shape[0] - 2 * r)
        for i in range(prev_start, min(start + 2 * r, padded.shape[0])):
            s += padded[i]
        for i in range(start, tail):
            for j in np.ndindex(data.shape[1:]):
                s[j] += padded[(i + 2 * r,) + j]
                prev = padded[(i,) + j]
                padded[(i,) + j] = s[j]
                s[j] -= prev
        for i in range(tail, stop):
            for j in np.ndindex(data.shape[1:]):
                prev = padded[(i,) + j]
                padded[(i,) + j] = s[j]
                s[j] -= prev
        prev_start = start
    for idx in np.ndindex(out.shape):
        out[idx] = padded[idx] / data.dtype.type(d)**K


@numba.jit(nopython=True, nogil=True)
def _box_gaussian_filter(data, sigma, out, passes=4):
    """Filter `data` with an approximate Gaussian filter.

    The filter is based on repeated filtering with a boxcar function. See
    [Get13]_ for details. It has finite support. Values outside the boundary
    are taken as zero.

    This function is not suitable when the input contains non-finite values,
    or very large variations in magnitude, as it internally computes a rolling
    sum. It also quantizes the requested sigma.

    .. [Get13] Pascal Getreuer, A Survey of Gaussian Convolution Algorithms,
       Image Processing On Line, 3 (2013), pp. 286-310.

    Parameters
    ----------
    data : ndarray
        Input data to filter (2D)
    sigma : ndarray
        Standard deviation of the Gaussian filter, per axis
    out : ndarray
        Output data, with the same shape as the input
    passes : int
        Number of boxcar filters to apply
    """
    if len(sigma) != data.ndim:
        raise ValueError('sigma has wrong number of elements')
    assert data.ndim == 2
    r = (0.5 * np.sqrt(12.0 * sigma**2 / passes + 1)).astype(np.int_)
    need_copy = True
    if r[0] > 0:
        # Process chunks of columns. See _sum_threshold for explanation.
        step = 256
        for i in range(0, data.shape[1], step):
            sub = slice(i, min(i + step, data.shape[1]))
            _box_gaussian_filter1d(data[:, sub], r[0], out[:, sub], passes)
        data = out       # Use out in next step
        need_copy = False
    if r[1] > 0:
        for i in range(data.shape[0]):
            _box_gaussian_filter1d(data[i], r[1], out[i], passes)
        need_copy = False
    if need_copy:
        out[:] = data[:]


@numba.jit(nopython=True, nogil=True)
def masked_gaussian_filter(data, flags, sigma, out, passes=4):
    """Filter an image using an approximate Gaussian filter.

    Some values may be flagged and are ignored. Values outside the grid are
    also treated as if flagged.

    See :func:`_box_gaussian_filter` for a number of caveats. The result may
    contain non-finite values where the finite support of the Gaussian
    approximation contains no values without flags.

    Parameters
    ----------
    data : ndarray, 2D
        Input data to filter
    flags : ndarray, bool
        True values correspond to elements of `data` to be ignored
    sigma : float or sequence of floats
        Standard deviation of the Gaussian filter, per axis
    out : ndarray, 2D
        Output array, with same shape as `data`
    passes : int
        Number of boxcar filters to apply
    """
    if data.shape != flags.shape:
        raise ValueError('shape mismatch between data and flags')
    if data.shape != out.shape:
        raise ValueError('shape mismatch between data and out')
    weight = np.empty_like(data)
    for idx in np.ndindex(data.shape):
        weight[idx] = not flags[idx]
        out[idx] = 0 if flags[idx] else data[idx]
    _box_gaussian_filter(weight, sigma, weight, passes=passes)
    _box_gaussian_filter(out, sigma, out, passes=passes)
    for idx in np.ndindex(out.shape):
        # Numeric instability can make out non-zero (but tiny) even
        # where filtered_weight is zero, which would make the ratio +/-inf
        # rather than NaN. Set to NaN explicitly in this case.
        if weight[idx] == 0:
            out[idx] = np.nan
        else:
            out[idx] /= weight[idx]


@numba.jit(nopython=True, nogil=True)
def _get_background2d(data, flags, iterations, spike_width, reject_threshold, freq_chunk_ends):
    """Determine a smooth background over a 2D array.

    This is done by iteratively convolving the data with elliptical Gaussians
    with linearly decreasing width from `iterations`*`spike_width` down to
    `spike width`. Outliers greater than `reject_threshold`*sigma from the
    background are masked on each iteration.

    Initial weights are set to zero at positions specified in `in_flags` if given.
    After the final iteration a final Gaussian smoothed background is computed
    and any stray NaNs in the background are interpolated in frequency (axis 1)
    for each timestamp (axis 0). The NaNs can appear when the the convolving
    Gaussian is completely covering masked data as the sum of convolved weights
    will be zero.

    Parameters
    ----------
    data : 2D ndarray, float
        The input data array to be smoothed, with shape (time, frequency).
    flags : 2D ndarray, boolean
        Flags corresponding to `data`
    iterations : int
        Number of iterations of Gaussian filtering
    spike_width : ndarray, float
        Two-element array containing the 1-sigma radius of the Gaussian filter
        in each axis.
    reject_threshold : float
        Number of standard deviations above which to flag data.
    freq_chunk_ends : ndarray, float
        Endpoints of intervals in which to compute noise estimates
        independently. This array must start with 0 and end of the number of
        channels, and be strictly increasing.
    """
    n_time, n_freq = data.shape
    flags = flags.copy()   # Gets modified
    background = np.empty_like(data)
    for extend_factor in range(iterations, 0, -1):
        sigma = extend_factor * spike_width
        masked_gaussian_filter(data, flags, sigma, background)
        for c in range(freq_chunk_ends.size - 1):
            sub = (slice(None, None), slice(freq_chunk_ends[c], freq_chunk_ends[c + 1]))
            sub_data = data[sub]
            sub_flags = flags[sub]
            # Convert background to an absolute value residual, in-place
            sub_residual = background[sub]
            for t in range(n_time):
                for f in range(sub_data.shape[1]):
                    sub_residual[t, f] = np.abs(sub_data[t, f] - sub_residual[t, f])
            threshold = _median_abs(sub_residual, sub_flags)
            threshold *= MAD_NORMAL * reject_threshold
            for t in range(n_time):
                for f in range(sub_data.shape[1]):
                    # sub_residual can contain NaNs, but only where the flags already apply
                    if sub_residual[t, f] > threshold:
                        sub_flags[t, f] = True
    # Compute final background
    masked_gaussian_filter(data, flags, spike_width, background)
    # Remove NaNs via linear interpolation
    _linearly_interpolate_nans(background)
    return background


@numba.jit(nopython=True, nogil=True)
def _convolve_flags(in_values, scale, threshold, out_flags, window):
    """Flag values with a threshold, and smear the flags.

    This is rolled into a single function for efficient implementation, but
    there are logically several steps:
    - For each value v in `in_values`, flag it if ``v * scale > threshold``.
    - Convolve the flags by a box filter of size `window`, expanding the width.
    - Logical OR these new flags into `out_flags`.
    """
    cum_size = in_values.shape[0] + 2 * window - 1
    # TODO: could preallocate this externally
    cum = np.empty((cum_size,) + (in_values.shape[1:]), np.uint32)     # Cumulative flagged values
    cum[:window] = 0
    for i in range(in_values.shape[0]):
        for j in np.ndindex(in_values.shape[1:]):
            flag = in_values[(i,) + j] * scale > threshold[(0,) + j]
            cum[(window + i,) + j] = cum[(window + i - 1,) + j] + flag
    # numba doesn't seem to fully support negative indices, hence
    # the addition of cum_size.
    cum[cum_size - (window - 1):] = cum[cum_size - window]
    for i in range(out_flags.shape[0]):
        for j in np.ndindex(out_flags.shape[1:]):
            out_flags[(i,) + j] |= (cum[(i + window,) + j] - cum[(i,) + j] != 0)


@numba.jit(nopython=True, nogil=True)
def _sum_threshold1d(input_data, input_flags, output_flags, windows, outlier_nsigma, rho, chunks):
    """Implement :func:`_sum_threshold`.

    It operates along the first axis.
    """
    for ci in range(chunks.size - 1):
        chunk_slice = slice(chunks[ci], chunks[ci + 1])
        chunk_data = input_data[chunk_slice]
        chunk_flags = input_flags[chunk_slice]

        # Get standard deviation using MAD and set up initial threshold
        threshold = _median_abs_axis0(chunk_data, chunk_flags)
        threshold_scale = outlier_nsigma * MAD_NORMAL
        for idx in np.ndindex(threshold.shape):
            if np.isnan(threshold[idx]):
                threshold[idx] = np.inf
            else:
                threshold[idx] *= threshold_scale

        padded_slice = slice(max(chunks[ci] - np.max(windows) + 1, 0),
                             min(chunks[ci + 1] + np.max(windows) - 1, input_data.size))
        padded_data = input_data[padded_slice]
        # TODO: can pre-allocate these outside the loop (but will need resizing)
        output_flags_pos = np.zeros(padded_data.shape, np.bool_)
        output_flags_neg = np.zeros(padded_data.shape, np.bool_)
        for window in windows:
            # The threshold for this iteration is calculated from the initial threshold
            # using the equation from Offringa (2010).
            tf = pow(rho, np.log2(window))
            # Get the thresholds
            thisthreshold = threshold / tf
            # Set already flagged values to be the +/- value of the
            # threshold if they are outside the threshold, and take
            # a cumulative sum.
            cum_data = np.empty((padded_data.shape[0] + 1,) + padded_data.shape[1:], np.float64)
            cum_data[0] = 0
            for i in range(padded_data.shape[0]):
                for j in np.ndindex(padded_data.shape[1:]):
                    idx = (i,) + j
                    clamped = padded_data[idx]
                    limit = thisthreshold[(0,) + j]
                    if output_flags_pos[idx] and clamped > limit:
                        clamped = limit
                    elif output_flags_neg[idx] and clamped < -limit:
                        clamped = -limit
                    cum_data[(i + 1,) + j] = cum_data[idx] + clamped
            # Calculate a rolling sum array from the data with the window for this iteration,
            # which is later scaled by rolliing_scale to give the rolling average.
            avgarray = cum_data[window:] - cum_data[:-window]
            rolling_scale = np.float32(1.0 / window)

            # Work out the flags from the average data above the current threshold,
            # convolve them, and combine with current flags.
            _convolve_flags(avgarray, rolling_scale, thisthreshold, output_flags_pos, window)

            # Work out the flags from the average data below the current threshold,
            # convolve them, and OR with current flags.
            _convolve_flags(avgarray, -rolling_scale, thisthreshold, output_flags_neg, window)

        # Extract just the portion of output_flags_pos/neg corresponding to the
        # chunk itself, without the padding
        rel_slice = slice(chunk_slice.start - padded_slice.start,
                          chunk_slice.stop - padded_slice.start)
        output_flags[chunk_slice] = output_flags_pos[rel_slice] | output_flags_neg[rel_slice]


@numba.jit(nopython=True, nogil=True)
def _sum_threshold(input_data, input_flags, axis, windows, outlier_nsigma, rho, chunks=None):
    """Apply the SumThreshold method along the given axis of `input_data`.

    Parameters
    ----------
    input_data : ndarray, real
        Deviations from the background. The implementation is optimised for
        2D (and does not currently work in 1D), but higher dimensions are
        supported.
    input_flags : ndarray, bool
        Input flags. Used as a mask when computing the initial
        standard deviations of the input data.
    axis : int
        The axis on which to apply the SumThreshold operation. In the current
        implementation, must be 0 or 1.
    windows : ndarray, int
        Window sizes to average data in each SumThreshold step
    outlier_nsigma : float
        Number of standard deviations at which to flag
    rho : float
        Parameter controlling the relationship between threshold and window size
    chunks : ndarray, int
        Boundaries between chunks in which each chunk has a separate noise
        estimation. This array must start with 0, be strictly increasing, and
        end with ``input_data.shape[axis]``.

    Returns
    -------
    output_flags : ndarray, bool
        The derived flags
    """
    if chunks is None:
        chunks = np.array([0, input_data.shape[axis]])
    output_flags = np.empty(input_data.shape, np.bool_)
    if axis < 0 or axis >= input_data.ndim:
        raise ValueError('axis is out of range')
    elif axis == 1:
        for i in range(input_data.shape[0]):
            _sum_threshold1d(input_data[i], input_flags[i], output_flags[i],
                             windows, outlier_nsigma, rho, chunks)
    elif axis == 0:
        # The operation is independent of the other dimensions, but we process
        # them in chunks to be cache friendly. The step size should be big
        # enough that whole cache lines are used (even if the alignment is
        # poor), but small enough that multiple rows fit in L1. This heuristic
        # value assumes a 2D input.
        step = 256
        for i in range(0, input_data.shape[1], step):
            sub = slice(i, min(i + step, input_data.shape[1]))
            _sum_threshold1d(input_data[:, sub], input_flags[:, sub], output_flags[:, sub],
                             windows, outlier_nsigma, rho, chunks)
    else:
        raise ValueError('axis must be 0 or 1')
    return output_flags


@numba.jit(nopython=True, nogil=True, cache=True)
def _get_flags_impl(
        in_data, in_flags, out_flags,
        outlier_nsigma, windows_time, windows_freq,
        background_reject, background_iterations,
        spike_width_time, spike_width_freq, time_extend, freq_extend,
        freq_chunk_ends, average_freq, flag_all_time_frac, flag_all_freq_frac,
        rho):
    n_time, n_freq, n_bl = in_data.shape
    # Average `in_data` in frequency. This is done unconditionally, because it
    # also does other useful steps (see the documentation).
    data, flags = _average_freq(in_data, in_flags, average_freq)

    # Output flags, in baseline-major order
    tmp_flags = np.empty((n_bl, n_time, n_freq), np.bool_)
    # Do operations independently per baseline.
    for bl in range(data.shape[0]):
        _get_baseline_flags(
            data[bl], flags[bl], tmp_flags[bl],
            outlier_nsigma, windows_time, windows_freq,
            background_reject, background_iterations,
            spike_width_time, spike_width_freq,
            time_extend, freq_extend,
            freq_chunk_ends, average_freq,
            flag_all_time_frac, flag_all_freq_frac,
            rho)

    # Transpose the output flags and explicitly flag nans from input
    for t in range(n_time):
        for f in range(n_freq):
            for bl in range(n_bl):
                out_flags[t, f, bl] = tmp_flags[bl, t, f] or np.isnan(in_data[t, f, bl])


@numba.jit(nopython=True, nogil=True)
def _combine_flags(spec_flags, time_flags, freq_flags, time_extend, out):
    """Combine several sources of flags and smear them in time.

    Parameters
    ----------
    spec_flags : 1D ndarray, bool
        Flags with shape (frequency)
    time_flags, freq_flags : 2D ndarray, bool
        Flags with shape (time, frequency)
    time_extend : int
        Width of the convolution kernel for time smearing (should be odd)
    out : 2D ndarray, bool
        Output flags
    """
    n_time, n_freq = time_flags.shape
    # Combine spec_flags, time_flags and freq_flags, and take a cumulative sum
    # along the time axis for the purposes of convolution.
    flag_sum = np.empty((n_time + 1, n_freq), time_extend.dtype)
    flag_sum[0] = 0
    for t in range(n_time):
        for f in range(n_freq):
            flag = spec_flags[0, f] or time_flags[t, f] or freq_flags[t, f]
            flag_sum[t + 1, f] = flag_sum[t, f] + flag
    # Difference the cumulative sums to get time-smeared flags.
    time_delta_lo = -(time_extend // 2)
    time_delta_hi = time_delta_lo + time_extend
    for t in range(n_time):
        # Rows to difference, clamping to the data limits
        t0 = max(t + time_delta_lo, 0)
        t1 = min(t + time_delta_hi, n_time)
        for f in range(n_freq):
            out[t, f] = (flag_sum[t0, f] != flag_sum[t1, f])


@numba.jit(nopython=True, nogil=True)
def _unaverage_freq(flags, freq_extend, average_freq, flag_all_time_frac, flag_all_freq_frac, out):
    """Perform final processing for a single baseline.

    1. Flags are replicated to undo the effect of frequency averaging.
    2. Flags are smeared, using a kernel of width `flag_all_freq_frac`.
    3. Times and frequencies where more than `flag_all_freq_frac` or
       `flag_all_time_frac` of the values are already flagged become fully
       flagged.
    """
    # Frequency replication and smearing
    n_time, n_freq = flags.shape
    orig_freq = out.shape[-1]
    flag_sum = np.empty(orig_freq + 1, np.int32)
    flag_sum_time = np.zeros(orig_freq, np.int32)
    freq_delta_lo = -(freq_extend // 2)
    freq_delta_hi = freq_delta_lo + freq_extend
    for t in range(n_time):
        flag_sum[0] = 0
        for f in range(orig_freq):
            flag_sum[f + 1] = flag_sum[f] + flags[t, f // average_freq]
        # Take differences of the cumulative sums to get smearing
        tot = 0
        for f in range(orig_freq):
            f0 = max(f + freq_delta_lo, 0)
            f1 = min(f + freq_delta_hi, orig_freq)
            flag = (flag_sum[f1] != flag_sum[f0])
            out[t, f] = flag
            tot += flag
            flag_sum_time[f] += flag
        # If too much is flagged, flag the entire time
        if tot > flag_all_freq_frac * orig_freq:
            out[t, :] = True

    # Flag all times if too much is flagged. This should be rare, so we
    # write in columns even though that's normally an unfriendly access
    # pattern.
    for f in range(orig_freq):
        if flag_sum_time[f] > n_time * flag_all_time_frac:
            out[:, f] = True


@numba.jit(nopython=True, nogil=True)
def _get_baseline_flags(
        data, flags, out_flags,
        outlier_nsigma, windows_time, windows_freq,
        background_reject, background_iterations,
        spike_width_time, spike_width_freq, time_extend, freq_extend,
        freq_chunk_ends, average_freq, flag_all_time_frac, flag_all_freq_frac,
        rho):
    """Compute flags for a single baseline.

    It is called after frequency averaging, but writes back un-averaged
    results.

    Parameters
    ----------
    data : ndarray, real
        Visibility magnitudes, with shape (time, frequency)
    flags : ndarray, bool
        User-input flags corresponding to `data`
    out_flags : ndarray, bool
        Returned flags (which will have more channels than `data` if
        `average_freq` is greater than 1).
    outlier_nsigma : float
        Number of standard deviations at which to flag
    windows_time : array, int
        Size of averaging windows to use in the SumThreshold method in time
    windows_freq : array, int
        Size of averaging windows to use in the SumThreshold method in frequency
    background_reject : float
        Number of sigma to reject outliers when backgrounding
    background_iterations : int
        Number of iterations to use when determining a smooth background, after each
        iteration data in excess of `background_reject`*`sigma` are masked
    spike_width_time : float
        Characteristic width in dumps to smooth over when backgrounding. This is
        the one-sigma width of the convolving Gaussian in axis 0.
    spike_width_freq : float
        Characteristic width in channels to smooth over when backgrounding. This is
        the one-sigma width of the convolving Gaussian in axis 1.
    time_extend : int
        Size of kernel in time to convolve with flags after detection
    freq_extend : int
        Size of kernel in frequency to convolve with flags after detection
    freq_chunk_ends : ndarray, float
        Endpoints of intervals in which to compute noise estimates
        independently. This array must start with 0 and end of the number of
        channels, and be strictly increasing.
    average_freq : int
        Number of channels to average frequency before flagging. Flags will be extended
        to the frequency shape of the input data before being returned
    flag_all_time_frac : float
        Fraction of data flagged above which to extend flags to all data in time axis.
    flag_all_freq_frac : float
        Fraction of data flagged above which to extend flags to all data in frequency axis.
    rho : float
        Parameter controlling the relationship between threshold and window size
    """
    n_time, n_freq = data.shape
    # Generate median spectrum, background it, and flag it
    spec_data, spec_flags = _time_median(data, flags)
    spec_background = _get_background2d(spec_data, spec_flags, background_iterations,
                                        np.array((0.0, spike_width_freq)),
                                        background_reject,
                                        freq_chunk_ends)
    spec_data -= spec_background
    spec_flags = _sum_threshold(spec_data, spec_flags, 1, windows_freq,
                                outlier_nsigma, rho, freq_chunk_ends)
    # Broadcast spectral flags to per-timestamp
    flags |= spec_flags

    # Get and subtract 2D background
    background = _get_background2d(data, flags, background_iterations,
                                   np.array((spike_width_time, spike_width_freq)),
                                   background_reject,
                                   freq_chunk_ends)
    data -= background
    # SumThreshold along time axis
    time_flags = _sum_threshold(data, flags, 0, windows_time,
                                outlier_nsigma, rho)
    # SumThreshold along frequency axis - with time flags in the input flags
    flags |= time_flags
    freq_flags = _sum_threshold(data, flags, 1, windows_freq,
                                outlier_nsigma, rho, freq_chunk_ends)

    # Combine flag sources and do time smearing. We overwrite 'flags' since the
    # previous result is no longer needed.
    _combine_flags(spec_flags, time_flags, freq_flags, time_extend, flags)

    _unaverage_freq(flags, freq_extend, average_freq,
                    flag_all_time_frac, flag_all_freq_frac, out_flags)


def _get_flags_mp(in_data, in_flags, flagger):    # noqa: D401
    """Callback function for ProcessPoolExecutor.

    It allocates its own storage for the output.
    """
    out_flags = np.empty_like(in_flags)
    flagger._get_flags(in_data, in_flags, out_flags)
    return out_flags


class SumThresholdFlagger:
    """Flagger that detects spikes in both frequency and time axes.

    It uses the SumThreshold method (Offringa, A., MNRAS, 405, 155-167, 2010).
    The full algorithm does the following:

        1. Average the data in the frequency dimension (axis 1) into bins of
           size `self.average_freq`
        2. Divide the data into overlapping sub-chunks in frequency which are
           backgrounded and thresholded independently
        3. Flag a 1d spectrum median filtered in time to get fainter contaminated
           channels.
        4. Derive a smooth 2d background through each chunk
        5. SumThreshold the background subtracted chunks in time and frequency
        6. Extend derived flags in time and frequency, via self.freq_extend and
           self.time_extend
        7. Extend flags to all times and frequencies in cases when more than
           a given fraction of samples are flagged (via `self.flag_all_time_frac` and
           `self.flag_all_freq_frac`)

    Parameters
    ----------
    outlier_nsigma : float
        Number of sigma to reject outliers when thresholding
    windows_time : array, int
        Size of averaging windows to use in the SumThreshold method in time
    windows_freq : array, int
        Size of averaging windows to use in the SumThreshold method in frequency
    background_reject : float
        Number of sigma to reject outliers when backgrounding
    background_iterations : int
        Number of iterations to use when determining a smooth background, after each
        iteration data in excess of `background_reject`*`sigma` are masked
    spike_width_time : float
        Characteristic width in dumps to smooth over when backgrounding. This is
        the one-sigma width of the convolving Gaussian in axis 0.
    spike_width_freq : float
        Characteristic width in channels to smooth over when backgrounding. This is
        the one-sigma width of the convolving Gaussian in axis 1.
    time_extend : int
        Size of kernel in time to convolve with flags after detection
    freq_extend : int
        Size of kernel in frequency to convolve with flags after detection
    freq_chunks : int
        Number of equal-sized chunks to independently flag in frequency. Smaller
        chunks will be less affected by variations in the band in the frequency domain.
    average_freq : int
        Number of channels to average frequency before flagging. Flags will be extended
        to the frequency shape of the input data before being returned
    flag_all_time_frac : float
        Fraction of data flagged above which to extend flags to all data in time axis.
    flag_all_freq_frac : float
        Fraction of data flagged above which to extend flags to all data in frequency axis.
    rho : float
        Falloff exponent for SumThreshold
    """

    def __init__(self, outlier_nsigma=4.5, windows_time=[1, 2, 4, 8],
                 windows_freq=[1, 2, 4, 8], background_reject=2.0, background_iterations=1,
                 spike_width_time=12.5, spike_width_freq=10.0, time_extend=3, freq_extend=3,
                 freq_chunks=10, average_freq=1, flag_all_time_frac=0.6, flag_all_freq_frac=0.8,
                 rho=1.3):
        self.outlier_nsigma = outlier_nsigma
        self.windows_time = windows_time
        # Scale the frequency windows, and remove possible duplicates
        windows_freq = np.ceil(np.array(windows_freq, dtype=np.float32) / average_freq)
        self.windows_freq = np.unique(windows_freq.astype(np.int_))
        self.background_reject = background_reject
        self.background_iterations = background_iterations
        self.spike_width_time = spike_width_time
        # Scale spike_width by average_freq
        self.spike_width_freq = spike_width_freq / average_freq
        self.time_extend = _as_min_dtype(time_extend)
        self.freq_extend = _as_min_dtype(freq_extend)
        self.freq_chunks = freq_chunks
        self.average_freq = _as_min_dtype(average_freq)
        self.flag_all_time_frac = flag_all_time_frac
        self.flag_all_freq_frac = flag_all_freq_frac
        self.rho = rho

    def _get_flags(self, in_data, in_flags, out_flags):
        """Flag a batch of baselines.

        The batches are doled out by :meth:`get_flags`, either to an executor
        pool or directly. The batching is important because it affects memory
        access patterns. The batch size should not be too large, as otherwise
        it will overload the cache.

        This function is the interface between Python code and numba code, and
        takes care of conditioning the parameters into a form that the numba
        code can consume. All the actual work is done in
        :func:`_get_flags_impl`.
        """
        averaged_channels = (in_data.shape[1] + self.average_freq - 1) // self.average_freq

        # Set up frequency chunks
        freq_chunk_ends = np.linspace(0, averaged_channels, self.freq_chunks + 1).astype(np.int_)

        # Clip the windows to the available time and frequency range
        windows_time = np.array([w for w in self.windows_time if w <= in_data.shape[1]], np.int_)
        windows_freq = np.array([w for w in self.windows_freq if w <= averaged_channels], np.int_)

        _get_flags_impl(
            in_data, in_flags, out_flags,
            self.outlier_nsigma, windows_time, windows_freq,
            self.background_reject, self.background_iterations,
            self.spike_width_time, self.spike_width_freq,
            self.time_extend, self.freq_extend,
            freq_chunk_ends, self.average_freq,
            self.flag_all_time_frac, self.flag_all_freq_frac,
            self.rho)

    def get_flags(self, data, flags, pool=None, chunk_size=None, is_multiprocess=None):
        """Compute flags in data array.

        It has optional input `flags` of the same shape that denote samples in
        `data` to ignore when backgrounding and deriving thresholds.

        This can run in parallel if given a
        :class:`concurrent.futures.Executor`. Performance is generally better
        with a :class:`~concurrent.futures.ThreadPoolExecutor`. While a
        :class:`~concurrent.futures.ProcessPoolExecutor` is supported, it is
        usually limited by the speed at which the data can be pickled and
        transferred to the other processes.

        Parameters
        ----------
        data : 3D array
            The input visibility data, in (time, frequency, baseline) order. It may
            also contain just the magnitudes.
        flags : 3D array, boolean
            Input flags.
        pool : :class:`concurrent.futures.Executor`, optional
            Worker pool for parallel computation. If not specified,
            computation will be done serially.
        chunk_size : int, optional
            Number of baselines to process at a time. If not specified,
            heuristics are used to pick a reasonable value. Values above 16
            give diminishing returns and much larger values may actually reduce
            performance. Power-of-two sizes are likely to perform best.
        is_multiprocess : bool, optional
            If `pool` behaves like
            :class:`concurrent.futures.ProcessPoolExecutor` (in particular, if
            it makes copies of its arguments) then this must be set to
            ``True`` to invoke a slower path that ensures that results are
            returned and reassembled. If unspecified, it defaults to true for
            :class:`concurrent.futures.ProcessPoolExecutor` and false for all
            other types. Thus, it only needs to be specified when using an
            object that isn't a :class:`concurrent.futures.ProcessPoolExecutor`
            but behaves like one.

        Returns
        -------
        out_flags : 3D array, boolean, same shape as `data`
            Derived flags (True=flagged)

        """
        if data.shape != flags.shape:
            raise ValueError('Shape mismatch')
        if data.ndim != 3:
            raise ValueError('data has wrong number of dimensions')
        out_flags = np.empty(flags.shape, np.bool_)

        n_bl = data.shape[-1]
        if not chunk_size:
            chunk_size = 16
            if pool is not None:
                # Make sure there is enough parallelism. There is no way to
                # query the number of workers in a pool, so we'll just assume
                # it is equal to cpu_count. We want at least 4 tasks per CPU
                # to avoid load imbalances.
                workers = multiprocessing.cpu_count()
                while chunk_size > 1 and chunk_size * workers * 4 > n_bl:
                    chunk_size //= 2
        if pool is not None and is_multiprocess is None:
            is_multiprocess = isinstance(pool, concurrent.futures.ProcessPoolExecutor)
        futures = []
        outputs = {}
        try:
            for i in range(0, n_bl, chunk_size):
                chunk_data = data[..., i : i + chunk_size]
                chunk_flags = flags[..., i : i + chunk_size]
                chunk_out = out_flags[..., i : i + chunk_size]
                if pool is not None and is_multiprocess:
                    future = pool.submit(_get_flags_mp, chunk_data, chunk_flags, self)
                    outputs[future] = chunk_out
                    futures.append(future)
                elif pool is not None:
                    futures.append(pool.submit(self._get_flags, chunk_data, chunk_flags, chunk_out))
                else:
                    self._get_flags(chunk_data, chunk_flags, chunk_out)
            # Wait for all the futures to complete, and raise any exception.
            # In multiprocessing mode, copy results back.
            for future in concurrent.futures.as_completed(futures):
                result = future.result()
                if is_multiprocess:
                    outputs[future][:] = result
            return out_flags
        finally:
            # If there's an exception, stop any work we can
            for future in futures:
                future.cancel()



################################################################################
# Copyright (c) 2014-2022, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for :mod:`katsdpsigproc.tune`."""

import threading
from unittest import mock
from typing import Any, Generator, Mapping, NoReturn

import pytest
import sqlite3

from katsdpsigproc import tune
from katsdpsigproc.abc import AbstractContext


def test_autotune_basic() -> None:
    received = []
    received_lock = threading.Lock()

    def generate(a, b):
        with received_lock:
            received.append((a, b))
        return lambda iters: a * b

    best = tune.autotune(generate, time_limit=0.001, a=[1, 2], b=[7, 3])
    # Autotuning is parallel, so we can't assert anything about the order
    assert sorted(received) == [(1, 3), (1, 7), (2, 3), (2, 7)]
    assert best == {'a': 1, 'b': 3}


def test_autotune_empty() -> None:
    with pytest.raises(ValueError):
        tune.autotune(lambda x, y: lambda iters: 0, x=[1, 2], y=[])


class CustomError(RuntimeError):
    pass


def test_autotune_some_raise() -> None:
    def generate(x):
        if x == 1:
            raise CustomError('x = 1')

        def measure(iters):
            if x == 3:
                raise CustomError('x = 3')
            return -x
        return measure
    best = tune.autotune(generate, x=[0, 1, 2, 3])
    assert best == {'x': 2}


def generate_raise(x: Any) -> NoReturn:
    raise CustomError(f'x = {x}')


def test_autotune_all_raise() -> None:
    with pytest.raises(CustomError, match=r'^x = 3$') as exc_info:
        tune.autotune(generate_raise, x=[1, 2, 3])
    # Check that the traceback refers to the original site, not
    # where it was re-raised
    assert exc_info.traceback[-1].name == 'generate_raise'


class TestAutotuner:
    """Tests for the `autotuner` decorator.

    We use mocking to substitute an in-memory database, which is made to
    persist for the test instead of being closed each time the decorator is
    called.
    """

    autotune_mock = mock.Mock()

    @pytest.fixture
    def conn(self) -> Generator[sqlite3.Connection, None, None]:
        conn = sqlite3.connect(':memory:')
        yield conn
        conn.close()

    @pytest.fixture(autouse=True)
    def _prepare_autotune_mock(self) -> Generator[None, None, None]:
        # The autotune method has to be a class method due to magic in the
        # autotuner decorator, and hence autotune_mock has to be at class
        # scope. But we want it reset for each test.
        self.autotune_mock.reset()
        self.autotune_mock.side_effect = None
        yield
        self.autotune_mock.reset()

    @classmethod
    @tune.autotuner(test={'a': 3, 'b': -1})
    def autotune(cls, context: AbstractContext, param: str) -> Mapping[str, Any]:
        return cls.autotune_mock(context, param)

    @classmethod
    @tune.autotuner(test={'a': 3, 'b': -1})
    def autotune_no_args(cls, context: AbstractContext) -> Any:
        return cls.autotune_mock(context)

    @mock.patch('katsdpsigproc.tune._close_db')
    @mock.patch('katsdpsigproc.tune._open_db')
    @mock.patch('katsdpsigproc.tune.KATSDPSIGPROC_TUNE_MATCH', 'exact')
    def test_autotune_match(self, open_db_mock: mock.Mock, close_db_mock: mock.Mock,
                            conn: sqlite3.Connection) -> None:
        """Test behaviour of both autotune matching options ('exact' and 'nearest')."""
        open_db_mock.return_value = conn
        context = mock.NonCallableMock()

        # Check that autotuning is only performed once for a particular
        # context+arg_parameters set
        tuning_1 = {'a': 1, 'b': 2}
        self.autotune_mock.return_value = tuning_1
        context.device.driver_version = 'mock version'
        context.device.platform_name = 'mock platform'
        context.device.name = 'mock device'
        ret = self.autotune(context, 'xyz')
        ret_2 = self.autotune(context, 'xyz')
        self.autotune_mock.assert_called_once_with(context, 'xyz')
        assert ret == ret_2 == tuning_1

        # Check that using a different context+arg_parameters set causes tuning to be
        # re-run. This also populates the database with another set
        tuning_2 = {'a': 3, 'b': 4}
        self.autotune_mock.return_value = tuning_2
        context.device.driver_version = 'another version'
        context.device.platform_name = 'another platform'
        context.device.name = 'another device'
        ret = self.autotune(context, 'xyz')
        assert ret == tuning_2

        # Now test incremental fallback on device  [ driver version | platform | name ]
        # when KATSDPSIGPROC_TUNE_MATCH is set to 'nearest'
        with mock.patch('katsdpsigproc.tune.KATSDPSIGPROC_TUNE_MATCH', 'nearest'):
            # autotuning re-runs if arg_parameters are different
            tuning_4 = {'a': 7, 'b': 8}
            self.autotune_mock.return_value = tuning_4
            context.device.driver_version = 'mock version'
            context.device.platform_name = 'mock platform'
            context.device.name = 'mock device'
            ret = self.autotune(context, 'zzz')
            assert ret == tuning_4

            # Re-run with different configurations making sure the cached values are returned if
            # they exist in any configuration, and that autotuning does not run.
            self.autotune_mock.side_effect = RuntimeError

            context.device.driver_version = 'abc'
            context.device.platform_name = 'mock platform'
            context.device.name = 'mock device'
            ret = self.autotune(context, 'xyz')
            assert ret == tuning_1

            context.device.driver_version = 'abc'
            context.device.platform_name = 'abc'
            context.device.name = 'another device'
            ret = self.autotune(context, 'xyz')
            assert ret == tuning_2

            context.device.driver_version = 'mock version'
            context.device.platform_name = 'mock platform'
            context.device.name = 'mock device'
            ret = self.autotune(context, 'xyz')
            assert ret == tuning_1

            # If there is no match in the database, any tuning (for that parameter set) will
            # suffice
            context.device.driver_version = 'x'
            context.device.platform_name = 'y'
            context.device.name = 'z'
            ret = self.autotune(context, 'xyz')
            assert ret == tuning_1 or ret == tuning_2

    @mock.patch('katsdpsigproc.tune._close_db')
    @mock.patch('katsdpsigproc.tune._open_db')
    def test_autotune_empty_arg(self, open_db_mock: mock.Mock, close_db_mock: mock.Mock,
                                conn: sqlite3.Connection) -> None:
        """Test behaviour of empty parameter sets."""
        open_db_mock.return_value = conn
        context = mock.NonCallableMock()

        tuning = {'a': 1, 'b': 2}
        self.autotune_mock.return_value = tuning
        context.device.driver_version = 'mock version'
        context.device.platform_name = 'mock platform'
        context.device.name = 'mock device'

        ret = self.autotune_no_args(context)
        assert ret == tuning



################################################################################
# Copyright (c) 2014, 2019, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""RFI flagging package.

The package consists of a number of algorithms for *backgrounding*,
*noise estimation* and *thresholding*. Backgrounding finds a smooth version of
the spectrum, with RFI removed. Noise estimation computes statistics on the
Thresholding applies to the difference between the original spectrum and the
background, and identifies the RFI as deviations that are too large to be
noise.

A backgrounding, a noise estimation and a thresholding algorithm are combined
in either :class:`host.FlaggerHost`, :class:`device.FlaggerDevice`, or
:class:`device.FlaggerHostFromDevice`.
"""

MAD_NORMAL = 1.4826
"""Ratio between `median absolute deviation`_ and standard deviation of a Gaussian distribution.

.. _median absolute deviation: https://en.wikipedia.org/wiki/Median_absolute_deviation
"""

# Automatically added by katversion
__version__ = '1.9.0'



import logging
import numpy as np
import katsdpsigproc.accel
from katsdpsigproc.accel import Operation, IOSlot, Dimension, build, roundup
import katsdpsigproc.tune

SOURCE = """
<%include file="/port.mako"/>

KERNEL void multiply(GLOBAL float *data, float scale) {
    data[get_global_id(0)] *= scale;
}
"""

class MultiplyTemplate:
    def __init__(self, context, tuning=None):
        if tuning is None:
            tuning = self.autotune(context)
        self.wgs = tuning['wgs']
        self.program = build(context, '', source=SOURCE)

    @classmethod
    @katsdpsigproc.tune.autotuner(test={'wgs': 32})
    def autotune(cls, context):
        queue = context.create_tuning_command_queue()
        size = 1048576

        def generate(wgs):
            fn = cls(context, {'wgs': wgs}).instantiate(queue, size, 1)
            fn.ensure_all_bound()
            fn.buffer('data').zero(queue)
            return katsdpsigproc.tune.make_measure(queue, fn)

        return katsdpsigproc.tune.autotune(generate, wgs=[32, 64, 128, 256])

    def instantiate(self, queue, size, scale):
        return Multiply(self, queue, size, scale)


class Multiply(Operation):
    def __init__(self, template, queue, size, scale):
        super().__init__(queue)
        self.template = template
        self.kernel = template.program.get_kernel('multiply')
        self.scale = np.float32(scale)
        self.slots['data'] = IOSlot((Dimension(size, self.template.wgs),), np.float32)

    def _run(self):
        data = self.buffer('data')
        self.command_queue.enqueue_kernel(
            self.kernel,
            [data.buffer, self.scale],
            global_size=(roundup(data.shape[0], self.template.wgs),),
            local_size=(self.template.wgs,)
        )


logging.basicConfig(level='DEBUG')
ctx = katsdpsigproc.accel.create_some_context()
queue = ctx.create_command_queue()
op_template = MultiplyTemplate(ctx)
op = op_template.instantiate(queue, 50, 3.0)
op.ensure_all_bound()
buf = op.buffer('data')
host = buf.empty_like()
host[:] = np.random.uniform(size=host.shape)
buf.set(queue, host)
op()
buf.get(queue, host)
print(host)



################################################################################
# Copyright (c) 2011-2022, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for :mod:`katsdpsigproc.fft`."""

import ctypes
from typing import Any, Tuple

import numpy as np
try:
    from numpy.typing import DTypeLike
except ImportError:
    DTypeLike = Any  # type: ignore
import pytest

from . import complex_normal
from katsdpsigproc.abc import AbstractContext, AbstractCommandQueue
from katsdpsigproc import fft


class TestCufft:
    """Test error handling in the :class:`._Cufft` helper class."""

    @pytest.fixture
    def cufft(self):
        return fft._Cufft()

    def test_invalid_plan(self, cufft: fft._Cufft) -> None:
        plan = cufft.cufftHandle(123)  # An invalid plan
        with pytest.raises(fft._Cufft.CufftError, match="CUFFT_INVALID_PLAN"):
            cufft.cufftSetAutoAllocation(plan, False)

    def test_unknown_error(self) -> None:
        with pytest.raises(fft._Cufft.CufftError, match="cuFFT error 0x7b"):
            raise fft._Cufft.CufftError(123)

    def test_bad_cenum(self, cufft: fft._Cufft) -> None:
        with pytest.raises(ctypes.ArgumentError):
            x = ctypes.c_longlong()
            work_size = ctypes.c_size_t()
            cufft.cufftMakePlanMany64(
                cufft.cufftHandle(), x, x, x, x, x, x, x, x, "bad", x, work_size
            )


@pytest.mark.cuda_only
class TestFft:
    @pytest.mark.parametrize(
        'shape, dtype, padded_shape_src, padded_shape_dest',
        [
            ((7, 16, 48), np.complex64, (7, 16, 48), (7, 18, 51)),
            ((7, 16, 48), np.complex128, (7, 16, 48), (7, 18, 51))
        ]
    )
    def test_c2c_forward(
        self,
        context: AbstractContext,
        command_queue: AbstractCommandQueue,
        shape: Tuple[int, ...],
        dtype: DTypeLike,
        padded_shape_src: Tuple[int, ...],
        padded_shape_dest: Tuple[int, ...]
    ):
        rs = np.random.RandomState(1)
        template = fft.FftTemplate(
            context, 2, shape, dtype, dtype, padded_shape_src, padded_shape_dest)
        fn = template.instantiate(command_queue, fft.FftMode.FORWARD)
        fn.ensure_all_bound()
        src = fn.buffer('src')
        dest = fn.buffer('dest')
        src_host = complex_normal(rs, size=src.shape).astype(dtype)
        src.set(command_queue, src_host)
        fn()
        expected = np.fft.fftn(src_host, axes=(-2, -1))
        tol = np.finfo(dtype).resolution * np.max(np.abs(expected))
        np.testing.assert_allclose(expected, dest.get(command_queue), atol=tol)

    @pytest.mark.parametrize(
        'shape, dtype, padded_shape_src, padded_shape_dest',
        [
            ((7, 16, 48), np.complex64, (7, 16, 48), (7, 18, 51)),
            ((7, 16, 48), np.complex128, (7, 16, 48), (7, 18, 51))
        ]
    )
    def test_c2c_inverse(
        self,
        context: AbstractContext,
        command_queue: AbstractCommandQueue,
        shape: Tuple[int, ...],
        dtype: DTypeLike,
        padded_shape_src: Tuple[int, ...],
        padded_shape_dest: Tuple[int, ...]
    ):
        rs = np.random.RandomState(1)
        template = fft.FftTemplate(
            context, 2, shape, dtype, dtype, padded_shape_src, padded_shape_dest)
        fn = template.instantiate(command_queue, fft.FftMode.INVERSE)
        fn.ensure_all_bound()
        src = fn.buffer('src')
        dest = fn.buffer('dest')
        src_host = complex_normal(rs, size=src.shape).astype(dtype)
        src.set(command_queue, src_host)
        fn()
        expected = np.fft.ifftn(src_host, axes=(-2, -1)) * (src.shape[-2] * src.shape[-1])
        tol = np.finfo(dtype).resolution * np.max(np.abs(expected))
        np.testing.assert_allclose(expected, dest.get(command_queue), atol=tol)

    @pytest.mark.parametrize(
        'shape, dtype_src, dtype_dest, padded_shape_src, padded_shape_dest',
        [
            ((3, 2, 16, 48), np.float32, np.complex64, (3, 2, 24, 64), (3, 2, 20, 27)),
            ((3, 2, 15, 47), np.float32, np.complex64, (3, 2, 23, 63), (3, 2, 17, 24)),
            ((3, 2, 16, 48), np.float64, np.complex128, (3, 2, 24, 64), (3, 2, 20, 27))
        ]
    )
    def test_r2c(
        self,
        context: AbstractContext,
        command_queue: AbstractCommandQueue,
        shape: Tuple[int, ...],
        dtype_src: DTypeLike,
        dtype_dest: DTypeLike,
        padded_shape_src: Tuple[int, ...],
        padded_shape_dest: Tuple[int, ...]
    ) -> None:
        rs = np.random.RandomState(1)
        template = fft.FftTemplate(
            context, 2, shape, dtype_src, dtype_dest, padded_shape_src, padded_shape_dest)
        fn = template.instantiate(command_queue, fft.FftMode.FORWARD)
        fn.ensure_all_bound()
        src = fn.buffer('src')
        dest = fn.buffer('dest')
        src_host = rs.standard_normal(src.shape).astype(dtype_src)
        src.set(command_queue, src_host)
        fn()
        expected = np.fft.rfftn(src_host, axes=(-2, -1))
        tol = np.finfo(dtype_src).resolution * np.max(np.abs(expected))
        np.testing.assert_allclose(expected, dest.get(command_queue), atol=tol)

    @pytest.mark.parametrize(
        'shape, dtype_src, dtype_dest, padded_shape_src, padded_shape_dest',
        [
            ((3, 2, 16, 48), np.complex64, np.float32, (3, 2, 20, 27), (3, 2, 24, 64)),
            ((3, 2, 15, 47), np.complex64, np.float32, (3, 2, 17, 24), (3, 2, 23, 63)),
            ((3, 2, 15, 47), np.complex128, np.float64, (3, 2, 17, 24), (3, 2, 23, 63))
        ]
    )
    def test_c2r(
        self,
        context: AbstractContext,
        command_queue: AbstractCommandQueue,
        shape: Tuple[int, ...],
        dtype_src: DTypeLike,
        dtype_dest: DTypeLike,
        padded_shape_src: Tuple[int, ...],
        padded_shape_dest: Tuple[int, ...]
    ) -> None:
        rs = np.random.RandomState(1)
        template = fft.FftTemplate(
            context, 2, shape, dtype_src, dtype_dest, padded_shape_src, padded_shape_dest)
        fn = template.instantiate(command_queue, fft.FftMode.INVERSE)
        fn.ensure_all_bound()
        src = fn.buffer('src')
        dest = fn.buffer('dest')
        signal = rs.standard_normal(shape).astype(np.float32) + 3
        src.set(command_queue, np.fft.rfftn(signal, axes=(-2, -1)).astype(dtype_src))
        fn()
        expected = signal * shape[2] * shape[3]  # CUFFT does unnormalised FFTs
        tol = np.finfo(dtype_src).resolution * np.max(np.abs(expected))
        np.testing.assert_allclose(expected, dest.get(command_queue), atol=tol)

    def test_wrong_direction(
        self,
        context: AbstractContext,
        command_queue: AbstractCommandQueue
    ) -> None:
        template = fft.FftTemplate(context, 1, (16,), np.float32, np.complex64, (16,), (9,),)
        with pytest.raises(ValueError, match='R2C transform must use FftMode.FORWARD'):
            template.instantiate(command_queue, fft.FftMode.INVERSE)
        template = fft.FftTemplate(context, 1, (16,), np.complex64, np.float32, (9,), (16,),)
        with pytest.raises(ValueError, match='C2R transform must use FftMode.INVERSE'):
            template.instantiate(command_queue, fft.FftMode.FORWARD)

    def test_bad_dtype_combination(
        self,
        context: AbstractContext,
        command_queue: AbstractCommandQueue
    ) -> None:
        with pytest.raises(ValueError, match='Invalid combination of dtypes'):
            fft.FftTemplate(context, 1, (16,), np.float32, np.complex128, (16,), (16,))
        with pytest.raises(ValueError, match='Invalid combination of dtypes'):
            fft.FftTemplate(context, 1, (16,), np.int32, np.int32, (16,), (16,))

    def test_length_mismatch(
        self,
        context: AbstractContext,
        command_queue: AbstractCommandQueue
    ) -> None:
        with pytest.raises(ValueError, match='padded_shape_src and shape'):
            fft.FftTemplate(context, 1, (16, 16), np.float32, np.complex64, (16,), (16, 16))
        with pytest.raises(ValueError, match='padded_shape_dest and shape'):
            fft.FftTemplate(context, 1, (16, 16), np.float32, np.complex64, (16, 16), (16,))

    def test_bad_padding(
        self,
        context: AbstractContext,
        command_queue: AbstractCommandQueue
    ) -> None:
        with pytest.raises(ValueError, match='Source must not be padded'):
            fft.FftTemplate(context, 1, (16, 16), np.complex64, np.complex64, (17, 16), (16, 16))
        with pytest.raises(ValueError, match='Destination must not be padded'):
            fft.FftTemplate(context, 1, (16, 16), np.complex64, np.complex64, (16, 16), (17, 16))



################################################################################
# Copyright (c) 2014-2020, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tools for autotuning algorithm parameters and caching the results.

Note that this is intended to apply only to parameters that affect
performance, such as work-group sizes, and not the outputs.

The design is that each computation class that implements autotuning will
provide a class method (typically called `autotune`) that takes a context and
user-supplied parameters (related to the problem rather than the
implementation), and returns a dictionary of values. A decorator is applied to
the autotuning method to cause the results to be cached.

The class may define an autotune_version class attribute to version the table.

At present, caching is implemented in an sqlite3 database. There is a table
corresponding to each autotuning method. The table has the following columns:

- `device_name`: string, name for the compute device
- `device_platform`: string, name for the compute device's platform
- `device_version`: version string for the driver
- `arg_*`: for each parameter passed to the autotuner
- `value_*`: for each value returned from the autotuner

The database is stored in the user cache directory.
"""

import itertools
import inspect
import os
import os.path
import enum
import time
import logging
import multiprocessing
import concurrent.futures
from typing import (
    Any,
    Callable,
    Dict,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    cast,
)

import appdirs
import sqlite3
import numpy as np
from decorator import decorator
from typing_extensions import Protocol

from .abc import AbstractContext, AbstractTuningCommandQueue


_logger = logging.getLogger(__name__)
_ScoreFunc = Callable[[int], float]
_T = TypeVar("_T")

KATSDPSIGPROC_TUNE_MATCH = os.getenv("KATSDPSIGPROC_TUNE_MATCH", "exact")
if KATSDPSIGPROC_TUNE_MATCH not in ["exact", "nearest"]:
    _logger.debug("KATSDPSIGPROC_TUNE_MATCH environment variable not one of [ exact | nearest ]: "
                  "setting to 'exact'.")
    KATSDPSIGPROC_TUNE_MATCH = "exact"


class _TuningFunc(Protocol):
    @property
    def __name__(self) -> str:
        ...

    def __call__(
        self, __cls: Type, __context: AbstractContext, *args: Any, **kwargs: Any
    ) -> Mapping[str, Any]:
        ...


def adapt_value(value: Any) -> Any:
    """Convert `value` to a type that can be used in sqlite3.

    This is not done through the sqlite3 adapter interface, because that is
    global rather than per-connection. This also only applies to lookup keys,
    not results, because it is not a symmetric relationship.
    """
    if isinstance(value, type) or isinstance(value, np.dtype):
        return repr(value)
    elif isinstance(value, enum.Enum):
        return value.name
    return value


def _db_keys(fn: _TuningFunc, args: Sequence, kwargs: Mapping) -> Dict[str, Any]:
    """Use the arguments passed to an autotuning function to generate a database key.

    .. seealso::

       :func:`autotuner`
    """
    # Extract the arguments passed to the wrapped function, by name
    sig = inspect.signature(fn)
    bound = sig.bind(*args, **kwargs)
    bound.apply_defaults()
    keys = {
        "arg_" + name: adapt_value(value)
        for (name, value) in itertools.islice(bound.arguments.items(), 2, None)
    }

    # Add information about the device
    device = args[1].device
    keys["device_name"] = device.name
    keys["device_platform"] = device.platform_name
    keys["device_version"] = device.driver_version
    return keys


def _query(
    conn: sqlite3.Connection, tablename: str, keys: Mapping[str, Any]
) -> Optional[sqlite3.Row]:
    query = f"SELECT * FROM {tablename}"
    query_args = []
    first = True
    for key, value in keys.items():
        if first:
            query += " WHERE"
        else:
            query += " AND"
        first = False
        query += f" {key}=?"
        query_args.append(value)
    cursor = conn.cursor()
    cursor.execute(query, query_args)
    row = cursor.fetchone()
    return row


def _fetch(
    conn: sqlite3.Connection, tablename: str, keys: Mapping[str, Any]
) -> Optional[Mapping[str, Any]]:
    """Fetch a cached record from the database.

    If the KATSDPSIGPROC_TUNE_MATCH environment variable is set to "nearest", and
    an exact record is not found, return the nearest match, by ignoring in turn
    device driver version, then device platform, then device name, or None
    otherwise (triggering autotuning).

    Parameters
    ----------
    conn
        Database connection
    tablename
        Name of the table to query
    keys
        Keys and values for the query
    """
    try:
        row = _query(conn, tablename, keys)
        if row is None:
            if KATSDPSIGPROC_TUNE_MATCH == "nearest":
                # Find the nearest autotune match -- if it exists -- by first ignoring
                # device driver version, then device platform, then device name.
                tune_keys = dict(keys)
                for key in ["device_version", "device_platform", "device_name"]:
                    _logger.debug("Retrying query by ignoring %s", key)
                    del tune_keys[key]
                    try:
                        row = _query(conn, tablename, tune_keys)
                    except sqlite3.Error:
                        _logger.debug("Query '%s' failed", exc_info=True)
                        pass
                    if row is not None:
                        break

        if row is not None:
            ans = {}
            for colname in row.keys():
                if colname.startswith("value_"):
                    # Truncate the 'value_' prefix
                    ans[colname[6:]] = row[colname]
            return ans
    except sqlite3.Error:
        # This could happen if the table does not exist yet
        _logger.debug("Query failed", exc_info=True)
        pass
    return None


def _create_table(
    conn: sqlite3.Connection,
    tablename: str,
    keys: Mapping[str, Any],
    values: Mapping[str, Any],
) -> None:
    command = f"CREATE TABLE IF NOT EXISTS {tablename} ("
    for name in itertools.chain(keys.keys(), values.keys()):
        command += name + " NOT NULL, "
    command += "PRIMARY KEY ({}) ON CONFLICT REPLACE)".format(", ".join(keys.keys()))
    conn.execute(command)


def _save(
    conn: sqlite3.Connection,
    tablename: str,
    keys: Mapping[str, Any],
    values: Mapping[str, Any],
) -> None:
    """ Write cached result into the table, creating it if it does not already exist."""
    _create_table(conn, tablename, keys, values)
    # Combine all fields
    entries = dict(keys)
    entries.update(values)
    command = "INSERT OR REPLACE INTO {}({}) VALUES ({})".format(
        tablename, ", ".join(entries.keys()), ", ".join(["?" for x in entries.keys()])
    )
    # Start transaction
    with conn:
        cursor = conn.cursor()
        cursor.execute(command, list(entries.values()))


def _open_db() -> sqlite3.Connection:
    cache_file = os.getenv("KATSDPSIGPROC_TUNE_DB")
    if cache_file is None:
        cache_dir = appdirs.user_cache_dir("katsdpsigproc", "ska-sa")
        os.makedirs(cache_dir, exist_ok=True)
        cache_file = os.path.join(cache_dir, "tuning.db")
    return sqlite3.connect(cache_file)


def _close_db(conn: sqlite3.Connection) -> None:
    """Close a database.

    This is split into a separate function for the benefit of testing, because
    the close member of the object itself is read-only and hence cannot be
    patched.
    """
    conn.close()


def autotuner_impl(
    test: Mapping[str, Any], fn: _TuningFunc, *args: Any, **kwargs: Any
) -> Mapping[str, Any]:
    """Implement :func:`autotuner`.

    It is split into a separate function so that mocks can patch it.
    """
    cls = args[0]
    classname = f"{cls.__module__}.{cls.__name__}.{fn.__name__}"
    tablename = (
        classname.replace(".", "_") + "__" + str(getattr(cls, "autotune_version", 0))
    )
    keys = _db_keys(fn, args, kwargs)

    conn = _open_db()
    conn.row_factory = sqlite3.Row
    try:
        ans = _fetch(conn, tablename, keys)
        if ans is None:
            # Nothing found in the database, so we need to tune now
            _logger.info("Performing autotuning for %s with key %s", classname, keys)
            ans = fn(*args, **kwargs)
            values = {"value_" + key: value for (key, value) in ans.items()}
            _save(conn, tablename, keys, values)
        else:
            _logger.debug("Autotuning cache hit for %s with key %s", classname, keys)
    finally:
        _close_db(conn)
    return ans


def autotuner(test: Mapping[str, Any]) -> Callable[[_T], _T]:
    r"""Decorate a function to make it an autotuning function that caches the result.

    The function must take a class and a context as the first two arguments.
    The remaining arguments form a cache key, along with properties of the
    device and the name of the function.

    Every argument to the function must have a name, which implies that the
    \*args construct may not be used.

    Parameters
    ----------
    test : dictionary
        A value that will be returned by :func:`stub_autotuner`.
    """

    @decorator
    def autotuner(fn: _TuningFunc, *args: Any, **kwargs: Any) -> Mapping[str, Any]:
        r"""Decorate a function to make it an autotuning function that caches the result.

        The function must take a class and a context as the first two
        arguments. The remaining arguments form a cache key, along with
        properties of the device and the name of the function.

        Every argument to the function must have a name, which implies that the
        \*args construct may not be used.
        """
        return autotuner_impl(test, fn, *args, **kwargs)

    # decorator module doesn't have type annotations, so we need to force it
    return cast(Callable[[_T], _T], autotuner)


def force_autotuner(
    test: Mapping[str, Any], fn: _TuningFunc, *args: Any, **kwargs: Any
) -> Mapping[str, Any]:
    """Drop-in replacement for :func:`autotuner_impl` that does not do any caching.

    It is intended to be used with a mocking framework.
    """
    return fn(*args, **kwargs)


def stub_autotuner(
    test: Mapping[str, Any], fn: _TuningFunc, *args: Any, **kwargs: Any
) -> Mapping[str, Any]:
    """Drop-in replacement for :func:`autotuner_impl` that does not do any tuning.

    It instead returns the provided value. It is intended to be
    used with a mocking framework.
    """
    return test


def make_measure(
    queue: AbstractTuningCommandQueue, function: Callable[[], None]
) -> _ScoreFunc:
    """Generate a measurement function.

    The result can be returned by the function passed to :func:`autotune`. It
    calls `function` (with no arguments) the appropriate number of times and
    returns the averaged elapsed time as measured by `queue`.
    """

    def measure(iters):
        queue.start_tuning()
        for i in range(iters):
            function()
        return queue.stop_tuning() / iters

    return measure


def autotune(
    generate: Callable[..., Optional[_ScoreFunc]],
    time_limit: float = 0.1,
    threads: Optional[int] = None,
    **kwargs: Any,
) -> Mapping[str, Any]:
    """Run a number of tuning experiments and find the optimal combination of parameters.

    Each argument is a iterable. The `generate` function is passed each
    element of the Cartesian product (by keyword), and returns a callable.
    This callable is passed an iteration count, and returns a score: lower is
    better. If either `generate` or the function it returns raises an
    exception, it is suppressed. Returns a dictionary with the best combination
    of values.

    Instead of a callable, the generate function may return `None` to skip a
    configuration that it knows will be unsuitable. Throwing an exception has
    essentially the same effect (and is used in code written before returning
    `None` was allowed), but returning `None` is preferred since an exception
    just clutters the log.

    The scoring function should not do a warmup pass: that is handled by this
    function.

    Parameters
    ----------
    generate
        function that creates a scoring function
    time_limit
        amount of time to spend testing each configuration (excluding setup time)
    threads
        number of parallel calls to `generate` to make at a time (defaults to
        CPU count)

    Raises
    ------
    Exception
        if every combination throws an exception, the last exception is re-raised.
    ValueError
        if the Cartesian product is empty
    """
    opts = itertools.product(*kwargs.values())
    best = None  # type: Optional[Mapping[str, Any]]
    best_score = None  # type: Optional[float]
    exc = None  # type: Optional[Exception]
    if threads is None:
        try:
            threads = multiprocessing.cpu_count()
        except NotImplementedError:
            threads = 1
    with concurrent.futures.ThreadPoolExecutor(threads) as thread_pool:
        while True:
            batch = list(itertools.islice(opts, threads))
            if not batch:
                break
            batch_keywords = [dict(zip(kwargs.keys(), opt)) for opt in batch]
            futures = [
                thread_pool.submit(generate, **keywords) for keywords in batch_keywords
            ]
            concurrent.futures.wait(futures)
            for keywords, future in zip(batch_keywords, futures):
                try:
                    measure = future.result()
                    if measure is None:
                        continue
                    # Do a warmup pass
                    measure(1)
                    # Do an initial timing pass
                    start = time.time()
                    measure(1)
                    # Take a max to prevent divide-by-zero in very fast cases
                    elapsed = max(time.time() - start, 1e-4)
                    # Guess how many iterations we can do in the allotted time
                    iters = max(3, int(time_limit / elapsed))
                    score = measure(iters)
                    _logger.debug(
                        "Configuration %s scored %f in %d iterations",
                        keywords,
                        score,
                        iters,
                    )
                    if best_score is None or score < best_score:
                        best = keywords
                        best_score = score
                except Exception as e:
                    exc = e
                    _logger.debug(
                        "Caught exception while testing configuration %s",
                        keywords,
                        exc_info=True,
                    )
    if best is None:
        if exc is not None:
            raise exc
        else:
            raise ValueError("No options to test")
    return best



import numpy as np
import katsdpsigproc.accel

ctx = katsdpsigproc.accel.create_some_context()
queue = ctx.create_command_queue()
dev = katsdpsigproc.accel.DeviceArray(ctx, (10,), np.float32)
host = dev.empty_like()
host[:] = 1
dev.set(queue, host)



################################################################################
# Copyright (c) 2014-2022, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Test RFI background estimation."""

import numpy as np
import pytest

from katsdpsigproc.rfi import host, device
from katsdpsigproc.abc import AbstractContext, AbstractCommandQueue
from .. import complex_normal


_vis = np.array([])
_vis_big = np.array([])
_flags = np.array([])
_flags_big = np.array([])


@pytest.fixture(autouse=True, scope="module")
def setup() -> None:
    global _vis, _vis_big, _flags, _flags_big
    shape = (417, 313)
    _vis = np.array([[1.25, 1.5j, 1.0, 2.0, -1.75, 2.0]]).T.astype(np.complex64)
    _flags = np.array([0, 0, 1, 0, 0, 4]).T.astype(np.uint8)
    # Use a fixed seed to make the test repeatable
    rs = np.random.RandomState(seed=1)
    _vis_big = complex_normal(rs, size=shape).astype(np.complex64)
    _flags_big = (rs.random_sample(shape) < 0.1).astype(np.uint8)
    # Ensure that in some cases the entire window is flagged. Also test with non-0/1
    # flag values.
    _flags_big[100:110, 0:100] = 4


class TestBackgroundMedianFilterHost:
    def setup_method(self) -> None:
        self.background = host.BackgroundMedianFilterHost(3)

    def test(self) -> None:
        out = self.background(_vis)
        ref = np.array([[-0.125, 0.25, -0.5, 0.25, -0.25, 0.125]]).T.astype(np.float32)
        np.testing.assert_equal(ref, out)

    def test_flags(self) -> None:
        out = self.background(_vis, _flags)
        ref = np.array([[-0.125, 0.125, 0.0, 0.125, -0.125, 0.0]]).T.astype(np.float32)
        np.testing.assert_equal(ref, out)


class TestBackgroundDevice:
    @pytest.fixture(params=[True, False])
    def amplitudes(self, request) -> bool:
        return request.param

    @pytest.fixture(params=[
        device.BackgroundFlags.NONE, device.BackgroundFlags.CHANNEL, device.BackgroundFlags.FULL
    ])
    def use_flags(self, request) -> device.BackgroundFlags:
        return request.param

    def test_result(self, context: AbstractContext, command_queue: AbstractCommandQueue,
                    amplitudes: bool, use_flags: device.BackgroundFlags) -> None:
        width = 5
        bg_device_template = device.BackgroundMedianFilterDeviceTemplate(
            context, width, amplitudes, use_flags)
        bg_host = bg_device_template.host_class(width, amplitudes)
        bg_device = device.BackgroundHostFromDevice(bg_device_template, command_queue)
        if amplitudes:
            vis = np.abs(_vis_big)
        else:
            vis = _vis_big
        if use_flags:
            full = use_flags == device.BackgroundFlags.FULL
            flags = _flags_big if full else _flags_big[:, 0]
            out_host = bg_host(vis, flags)
            out_device = bg_device(vis, flags)
        else:
            out_host = bg_host(vis)
            out_device = bg_device(vis)
        # Uses an abs tolerance because backgrounding subtracts nearby values
        np.testing.assert_allclose(out_host, out_device, atol=1e-6)

    @pytest.mark.force_autotune
    def test_autotune(self, context: AbstractContext,
                      amplitudes: bool, use_flags: device.BackgroundFlags) -> None:
        device.BackgroundMedianFilterDeviceTemplate(context, 5, amplitudes, use_flags)



################################################################################
# Copyright (c) 2014-2021, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Abstraction layer over PyOpenCL.

It implements the abstract interfaces defined by :mod:`katsdpsigproc.abc`.
"""

from typing import List, Tuple, Sequence, Optional, Callable, Type, TypeVar, Union, Any
from types import TracebackType

import pyopencl
import pyopencl.array
import numpy as np
try:
    from numpy.typing import DTypeLike
except ImportError:
    DTypeLike = Any     # type: ignore

from .abc import (AbstractProgram, AbstractKernel, AbstractDevice, AbstractContext,
                  AbstractEvent, AbstractCommandQueue, AbstractTuningCommandQueue)


_T = TypeVar('_T')
_D = TypeVar('_D', bound='Device')
_AnyBuffer = Union['_DummyArray', pyopencl.array.Array]


class _DummyArray:
    """Trivial dummy for :class:`pyopencl.array.Array`, that just has a `data` attribute."""

    def __init__(self, data: pyopencl.Buffer) -> None:
        self.data = data


class _PinnedAMD(np.ndarray):
    """Pinned memory abstraction for AMD GPUs.

    Copies to the device
    are done by unmapping the buffer, enqueuing the copy, and remapping
    the buffer. This is based on AMD's optimization guide.
    """

    _mapping: pyopencl.MemoryMap = None
    _buffer: pyopencl.Buffer
    _array: _DummyArray

    def __new__(cls, context: 'Context', queue: 'CommandQueue',
                shape: Tuple[int, ...], dtype: DTypeLike) -> '_PinnedAMD':
        dtype = np.dtype(dtype)
        n_bytes = int(np.prod(shape)) * dtype.itemsize
        # Do not add READ or WRITE to the flags: doing so seems to cause AMD
        # drivers to allocate GPU memory.
        buffer = pyopencl.Buffer(
            context,
            pyopencl.mem_flags.ALLOC_HOST_PTR,
            n_bytes)
        array = pyopencl.enqueue_map_buffer(
            queue,
            buffer,
            pyopencl.map_flags.READ | pyopencl.map_flags.WRITE,
            0, shape, dtype)[0]
        mapping = array.base               # type: pyopencl.MemoryMap
        array = array.view(_PinnedAMD)
        array._buffer = buffer
        array._mapping = mapping
        array._array = _DummyArray(buffer)
        return array

    def _enqueue_transfer(self, queue: 'CommandQueue', blocking: bool,
                          transfer: Callable[[], None]) -> None:
        if not hasattr(self, '_buffer'):
            raise ValueError('cannot copy to/from a _PinnedAMD view')
        self._mapping.release(queue._pyopencl_command_queue)
        transfer()
        # The shape and dtype don't matter here as long as we map the entire buffer,
        # because we're going to throw away the array wrapper and just keep the
        # MemoryMap object.
        array = pyopencl.enqueue_map_buffer(
            queue._pyopencl_command_queue,
            self._buffer, pyopencl.map_flags.READ | pyopencl.map_flags.WRITE,
            0, (self._buffer.size,), np.uint8, is_blocking=blocking)[0]
        self._mapping = array.base

    def enqueue_write_buffer(self, queue: 'CommandQueue', buffer: pyopencl.array.Array,
                             blocking: bool) -> None:
        """Enqueue a copy from the host buffer to device memory.

        The host memory must not be touched while the copy is in progress.

        Parameters
        ----------
        queue
            Queue for the copy
        buffer
            Target buffer
        blocking
            If `True`, wait for the copy to complete before returning
        """
        def transfer():
            queue.enqueue_copy_buffer(self._array, buffer)
        self._enqueue_transfer(queue, blocking, transfer)

    def enqueue_read_buffer(self, queue: 'CommandQueue', buffer: pyopencl.array.Array,
                            blocking: bool) -> None:
        """Enqueue a copy to the host buffer from device memory.

        The host memory must not be touched while the copy is in progress.

        Parameters
        ----------
        queue
            Queue for the copy
        buffer
            Source buffer
        blocking
            If `True`, wait for the copy to complete before returning
        """
        def transfer():
            queue.enqueue_copy_buffer(buffer, self._array)
        self._enqueue_transfer(queue, blocking, transfer)

    def enqueue_write_buffer_rect(
            self, queue: 'CommandQueue', buffer: pyopencl.array.Array,
            buffer_origin: int, data_origin: int, shape: Sequence[int],
            buffer_strides: Sequence[int], data_strides: Sequence[int],
            blocking: bool = True) -> None:
        def transfer():
            queue.enqueue_copy_buffer_rect(
                self._array, buffer, data_origin, buffer_origin, shape,
                data_strides, buffer_strides)
        self._enqueue_transfer(queue, blocking, transfer)

    def enqueue_read_buffer_rect(
            self, queue: 'CommandQueue', buffer: pyopencl.array.Array,
            buffer_origin: int, data_origin: int, shape: Sequence[int],
            buffer_strides: Sequence[int], data_strides: Sequence[int],
            blocking: bool = True) -> None:
        def transfer():
            queue.enqueue_copy_buffer_rect(
                buffer, self._array, buffer_origin, data_origin, shape,
                buffer_strides, data_strides)
        self._enqueue_transfer(queue, blocking, transfer)


class Program(AbstractProgram['Kernel']):
    def __init__(self, pyopencl_program: pyopencl.Program) -> None:
        self._pyopencl_program = pyopencl_program

    def get_kernel(self, name: str) -> 'Kernel':
        return Kernel(self, name)


class Kernel(AbstractKernel[Program]):
    def __init__(self, program: Program, name: str) -> None:
        self._pyopencl_kernel = pyopencl.Kernel(program._pyopencl_program, name)


class Event(AbstractEvent):
    def __init__(self, pyopencl_event: pyopencl.Event) -> None:
        self._pyopencl_event = pyopencl_event

    def wait(self) -> None:
        self._pyopencl_event.wait()

    def time_since(self, prior_event: 'Event') -> float:
        prior_event.wait()
        self.wait()
        return 1e-9 * (self._pyopencl_event.profile.start
                       - prior_event._pyopencl_event.profile.end)

    def time_till(self, next_event: 'Event') -> float:
        return next_event.time_since(self)


class Device(AbstractDevice['Context']):
    """Abstraction of an OpenCL device."""

    def __init__(self, pyopencl_device: pyopencl.Device) -> None:
        self._pyopencl_device = pyopencl_device

    def make_context(self) -> 'Context':
        return Context(pyopencl.Context([self._pyopencl_device]))

    @property
    def name(self) -> str:
        return self._pyopencl_device.name

    @property
    def platform_name(self) -> str:
        return self._pyopencl_device.platform.name

    @property
    def driver_version(self) -> str:
        return self._pyopencl_device.driver_version

    @property
    def is_cuda(self) -> bool:
        return False

    @property
    def is_gpu(self) -> bool:
        return self._pyopencl_device.type & pyopencl.device_type.GPU

    @property
    def is_accelerator(self) -> bool:
        return self._pyopencl_device.type & pyopencl.device_type.ACCELERATOR

    @property
    def is_cpu(self) -> bool:
        return self._pyopencl_device.type & pyopencl.device_type.CPU

    @property
    def simd_group_size(self) -> int:
        return pyopencl.characterize.get_simd_group_size(self._pyopencl_device, 4)

    @classmethod
    def _get_platforms(cls) -> List[pyopencl.Platform]:
        """Return all platforms."""
        try:
            return pyopencl.get_platforms()
        except pyopencl.LogicError:
            # OpenCL considers it an error if there are no platforms
            # available.
            return []

    @classmethod
    def get_devices(cls: Type[_D]) -> Sequence[_D]:
        ans = []
        for platform in cls._get_platforms():
            for device in platform.get_devices():
                ans.append(cls(device))
        return ans

    @classmethod
    def get_devices_by_platform(cls: Type[_D]) -> Sequence[Sequence[_D]]:
        ans = []
        for platform in cls._get_platforms():
            ans.append([cls(device) for device in platform.get_devices()])
        return ans


class Context(AbstractContext[pyopencl.array.Array, pyopencl.Buffer, None,
                              Device, Program, 'CommandQueue', 'TuningCommandQueue']):
    """Abstraction of an OpenCL context."""

    def __init__(self, pyopencl_context: pyopencl.Context) -> None:
        self._pyopencl_context = pyopencl_context
        device = pyopencl_context.devices[0]
        self._internal_queue = pyopencl.CommandQueue(pyopencl_context, device)
        self._force_pinned_amd = False   # Used for unit tests

    @property
    def device(self) -> Device:
        return Device(self._pyopencl_context.devices[0])

    def compile(self, source: str, extra_flags: Optional[List[str]] = None) -> Program:
        # source is passed through str because it might arrive as Unicode,
        # triggering a warning.
        # TODO: can remove this now that it's Python 3-only.
        program = pyopencl.Program(self._pyopencl_context, str(source))
        program.build(extra_flags if extra_flags is not None else [])
        return Program(program)

    def allocate_raw(self, n_bytes: int) -> pyopencl.Buffer:
        return pyopencl.Buffer(self._pyopencl_context, pyopencl.mem_flags.READ_WRITE, n_bytes)

    def allocate(self, shape: Tuple[int, ...], dtype: DTypeLike,
                 raw: Optional[pyopencl.Buffer] = None) -> pyopencl.array.Array:
        return pyopencl.array.Array(self._pyopencl_context, shape, dtype, data=raw)

    def allocate_pinned(self, shape: Tuple[int, ...], dtype: DTypeLike) -> np.ndarray:
        dtype = np.dtype(dtype)
        device = self._pyopencl_context.devices[0]
        if (self._force_pinned_amd
            or ((device.type & pyopencl.device_type.GPU)
                and device.platform.name == 'AMD Accelerated Parallel Processing')):
            return _PinnedAMD(self._pyopencl_context, self._internal_queue, shape, dtype)
        elif device.platform.name == 'NVIDIA CUDA':
            # Based on NVIDIA's recommendation: create a device buffer and
            # leave it mapped permanently.
            n_bytes = int(np.prod(shape)) * dtype.itemsize
            buf = pyopencl.Buffer(
                self._pyopencl_context,
                pyopencl.mem_flags.ALLOC_HOST_PTR | pyopencl.mem_flags.READ_ONLY,
                n_bytes)
            return pyopencl.enqueue_map_buffer(
                self._internal_queue,
                buf,
                pyopencl.map_flags.READ | pyopencl.map_flags.WRITE,
                0, shape, dtype)[0]
        else:
            return np.empty(shape, dtype)

    def allocate_svm_raw(self, n_bytes: int) -> None:
        raise NotImplementedError("PyOpenCL does not support OpenCL Shared Virtual Memory")

    def allocate_svm(self, shape: Tuple[int, ...], dtype: DTypeLike,
                     raw: Optional[None] = None) -> np.ndarray:
        raise NotImplementedError("PyOpenCL does not support OpenCL Shared Virtual Memory")

    def create_command_queue(self, profile: bool = False) -> 'CommandQueue':
        return CommandQueue(self, profile=profile)

    def create_tuning_command_queue(self) -> 'TuningCommandQueue':
        return TuningCommandQueue(self)

    def __enter__(self: _T) -> _T:
        return self

    def __exit__(self,
                 exc_type: Optional[Type[BaseException]],
                 exc_val: Optional[BaseException],
                 exc_tb: Optional[TracebackType]) -> None:
        pass


class CommandQueue(AbstractCommandQueue[pyopencl.array.Array, Context, Event, Kernel]):
    """Abstraction of a command queue.

    If no existing command queue is passed to the constructor, a new one is
    created.

    Parameters
    ----------
    context
        Context owning the queue
    pyopencl_command_queue
        Existing command queue to wrap
    profile
        If true and `pyopencl_command_queue` is omitted, enabling profiling
        (timing) on the queue.
    """

    def __init__(self, context: Context,
                 pyopencl_command_queue: Optional[pyopencl.CommandQueue] = None,
                 profile: bool = False) -> None:
        self.context = context
        if pyopencl_command_queue is None:
            pyopencl_device = context._pyopencl_context.devices[0]
            properties = 0
            if profile:
                properties |= pyopencl.command_queue_properties.PROFILING_ENABLE
            pyopencl_command_queue = pyopencl.CommandQueue(
                context._pyopencl_context, pyopencl_device,
                properties)
        self._pyopencl_command_queue = pyopencl_command_queue

    def enqueue_read_buffer(self, buffer: pyopencl.array.Array, data: Any,
                            blocking: bool = True) -> None:
        if hasattr(data, 'enqueue_read_buffer'):
            data.enqueue_read_buffer(self, buffer, blocking)
        else:
            buffer.get(ary=data, queue=self._pyopencl_command_queue, async_=not blocking)

    def enqueue_write_buffer(self, buffer: pyopencl.array.Array, data: Any,
                             blocking: bool = True) -> None:
        if hasattr(data, 'enqueue_write_buffer'):
            data.enqueue_write_buffer(self, buffer, blocking)
        else:
            buffer.set(ary=data, queue=self._pyopencl_command_queue, async_=not blocking)

    def enqueue_copy_buffer(
            self, src_buffer: _AnyBuffer, dest_buffer: _AnyBuffer) -> None:
        pyopencl.enqueue_copy(
            self._pyopencl_command_queue,
            src=src_buffer.data, dest=dest_buffer.data)

    def enqueue_copy_buffer_rect(
            self, src_buffer: _AnyBuffer, dest_buffer: _AnyBuffer,
            src_origin: int, dest_origin: int,
            shape: Sequence[int], src_strides: Sequence[int], dest_strides: Sequence[int]) -> None:
        assert src_strides[0] == 1
        assert dest_strides[0] == 1
        pyopencl.enqueue_copy(
            self._pyopencl_command_queue,
            src=src_buffer.data, dest=dest_buffer.data,
            src_origin=(src_origin,), dst_origin=(dest_origin,),
            region=shape,
            src_pitches=src_strides[1:], dst_pitches=dest_strides[1:])

    def enqueue_read_buffer_rect(
            self, buffer: pyopencl.array.Array, data: Any,
            buffer_origin: int, data_origin: int, shape: Sequence[int],
            buffer_strides: Sequence[int], data_strides: Sequence[int],
            blocking: bool = True) -> None:
        assert buffer_strides[0] == 1
        assert data_strides[0] == 1
        if hasattr(data, 'enqueue_read_buffer_rect'):
            data.enqueue_read_buffer_rect(self, buffer, buffer_origin, data_origin, shape,
                                          buffer_strides, data_strides, blocking)
        else:
            pyopencl.enqueue_copy(
                self._pyopencl_command_queue,
                src=buffer.data, dest=data,
                host_origin=(data_origin,), buffer_origin=(buffer_origin,),
                region=shape,
                host_pitches=data_strides[1:], buffer_pitches=buffer_strides[1:])

    def enqueue_write_buffer_rect(
            self, buffer: pyopencl.array.Array, data: Any,
            buffer_origin: int, data_origin: int, shape: Sequence[int],
            buffer_strides: Sequence[int], data_strides: Sequence[int],
            blocking: bool = True) -> None:
        assert buffer_strides[0] == 1
        assert data_strides[0] == 1
        if hasattr(data, 'enqueue_write_buffer_rect'):
            data.enqueue_write_buffer_rect(self, buffer, buffer_origin, data_origin, shape,
                                           buffer_strides, data_strides, blocking)
        else:
            pyopencl.enqueue_copy(
                self._pyopencl_command_queue,
                src=data, dest=buffer.data,
                host_origin=(data_origin,), buffer_origin=(buffer_origin,),
                region=shape,
                host_pitches=data_strides[1:], buffer_pitches=buffer_strides[1:])

    def enqueue_zero_buffer(self, buffer: pyopencl.array.Array) -> None:
        pyopencl.enqueue_fill_buffer(self._pyopencl_command_queue, buffer.data,
                                     np.uint8(0), 0, buffer.data.get_info(pyopencl.mem_info.SIZE))

    @classmethod
    def _raw_arg(cls, arg: Any) -> Any:
        if isinstance(arg, pyopencl.array.Array):
            return arg.data
        else:
            return arg

    def _enqueue_kernel(self, kernel: Kernel, args: Sequence[Any],
                        global_size: Tuple[int, ...],
                        local_size: Tuple[int, ...]) -> pyopencl.Event:
        """Enqueue a kernel to the command queue.

        This version returns the OpenCL event object. Refer to
        :meth:`enqueue_queue` for the public interface.
        """
        # PyOpenCL doesn't allow Array objects to be passed through
        args = [self._raw_arg(x) for x in args]
        # OpenCL allows local_size to be None, but this is not portable to
        # CUDA
        assert local_size is not None
        assert len(local_size) == len(global_size)
        return kernel._pyopencl_kernel(self._pyopencl_command_queue,
                                       global_size, local_size, *args)

    def enqueue_kernel(self, kernel: Kernel, args: Sequence[Any],
                       global_size: Tuple[int, ...], local_size: Tuple[int, ...]) -> None:
        self._enqueue_kernel(kernel, args, global_size, local_size)

    def enqueue_marker(self) -> Event:
        return Event(pyopencl.enqueue_marker(self._pyopencl_command_queue))

    def enqueue_wait_for_events(self, events: Sequence[Event]) -> None:
        # OpenCL has some odd semantics for an empty wait list, hence the check
        if events:
            pyopencl.enqueue_barrier(self._pyopencl_command_queue,
                                     [x._pyopencl_event for x in events])

    def flush(self) -> None:
        self._pyopencl_command_queue.flush()

    def finish(self) -> None:
        self._pyopencl_command_queue.finish()


class TuningCommandQueue(CommandQueue,
                         AbstractTuningCommandQueue[pyopencl.array.Array, Context, Event, Kernel]):
    """Command queue with extra facilities for autotuning.

    It keeps track of kernels that are enqueued since the last call to
    :meth:`start_tuning`, and reports the total time they consume when
    :meth:`stop_tuning` is called.
    """

    def __init__(self, *args, **kwargs) -> None:
        kwargs['profile'] = True
        super().__init__(*args, **kwargs)
        self.is_tuning = False
        self.events = []        # type: List[pyopencl.Event]

    def start_tuning(self) -> None:
        self.is_tuning = True
        self.events = []

    def enqueue_kernel(self, kernel: Kernel, args: Sequence[Any],
                       global_size: Tuple[int, ...], local_size: Tuple[int, ...]) -> None:
        if not self.is_tuning:
            super().enqueue_kernel(kernel, args, global_size, local_size)
        else:
            event = self._enqueue_kernel(kernel, args, global_size, local_size)
            self.events.append(event)

    def stop_tuning(self) -> float:
        self.finish()
        elapsed = 0.0
        if self.events:
            start = self.events[0].profile.start
            end = self.events[-1].profile.end
            elapsed = (end - start) * 1e-9
        self.is_tuning = False
        self.events = []
        return elapsed



################################################################################
# Copyright (c) 2014-2021, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""RFI flagging algorithms that run on an accelerator, using OpenCL or CUDA.

The noise estimation and thresholding functions may take data in either
channel-major or baseline-major order (the flags are emitted in the same
order). In the former case, the `transposed` member is `False`, otherwise it is
`True`. The flagger classes automatically detect this and apply a transposition
kernel at the appropriate point.
"""

import enum
from abc import ABC, abstractmethod
from typing import List, Tuple, Optional, Mapping, Union, Callable, Type, Any

import numpy as np

from .. import accel
from .. import tune
from .. import transpose
from ..abc import AbstractContext, AbstractCommandQueue
from ..accel import DeviceArray, AbstractAllocator
from . import host


_THRESHOLD_SUM_DEFAULT_THRESHOLD_FALLOFF = 1.2


class BackgroundFlags(enum.Enum):
    NONE = 0
    CHANNEL = 1
    FULL = 2

    def __bool__(self):
        return self != BackgroundFlags.NONE


class AbstractBackgroundDevice(accel.Operation):
    pass


class AbstractBackgroundDeviceTemplate(ABC):
    use_flags = None    # type: BackgroundFlags
    context = None      # type: AbstractContext
    host_class = None   # type: Type[host.AbstractBackgroundHost]

    @abstractmethod
    def instantiate(self, command_queue: AbstractCommandQueue, channels: int, baselines: int,
                    allocator: Optional[AbstractAllocator] = None) -> AbstractBackgroundDevice:
        """Create an instance."""


class AbstractNoiseEstDevice(accel.Operation):
    transposed = None   # type: bool


class AbstractNoiseEstDeviceTemplate(ABC):
    transposed = None   # type: bool
    context = None      # type: AbstractContext
    host_class = None   # type: Type[host.AbstractNoiseEstHost]

    @abstractmethod
    def instantiate(self, command_queue: AbstractCommandQueue, channels: int, baselines: int,
                    allocator: Optional[AbstractAllocator] = None) -> AbstractNoiseEstDevice:
        """Create an instance."""


class AbstractThresholdDevice(accel.Operation):
    transposed = None   # type: bool


class AbstractThresholdDeviceTemplate(ABC):
    transposed = None   # type: bool
    context = None      # type: AbstractContext
    host_class = None   # type: Type[host.AbstractThresholdHost]

    @abstractmethod
    def instantiate(self, command_queue: AbstractCommandQueue, channels: int, baselines: int,
                    n_sigma: float,
                    *, allocator: Optional[AbstractAllocator] = None) -> AbstractThresholdDevice:
        """Create an instance."""
        # allocator is marked as keyword-only in this base class because
        # concrete classes may have extra parameters before it.


class BackgroundHostFromDevice(host.AbstractBackgroundHost):
    """Wrap a device-side background template to present the host interface."""

    def __init__(self, template: AbstractBackgroundDeviceTemplate,
                 command_queue: AbstractCommandQueue) -> None:
        self.template = template
        self.command_queue = command_queue

    def __call__(self, vis: np.ndarray, flags: Optional[np.ndarray] = None) -> np.ndarray:
        if flags is not None and not self.template.use_flags:
            raise TypeError("flags were provided but not included in the template")
        if flags is None and self.template.use_flags:
            raise TypeError("flags were expected but not provided")
        (channels, baselines) = vis.shape
        fn = self.template.instantiate(self.command_queue, channels, baselines)
        # Trigger allocations
        fn.ensure_all_bound()
        fn.buffer('vis').set(self.command_queue, vis)
        if flags is not None:
            fn.buffer('flags').set(self.command_queue, flags)
        # Do the computation
        fn()
        return fn.buffer('deviations').get(self.command_queue)


class BackgroundMedianFilterDeviceTemplate(AbstractBackgroundDeviceTemplate):
    """Device algorithm that applies a median filter to each baseline (in amplitude).

    It is the same algorithm as :class:`host.BackgroundMedianFilterHost`, but
    may give slightly different results due to rounding errors when computing
    complex magnitude.

    Parameters
    ----------
    context
        Context for which kernels will be compiled
    width
        The kernel width (must be odd)
    is_amplitude
        If ``True``, the inputs are amplitudes rather than complex visibilities
    use_flags
        Specify that flags are taken as input to indicate bad data that should
        not contribute to the backgrounding. The legal values are

        - ``FULL``: takes flags with same shape as visibilities
        - ``CHANNEL``: takes a single flag per channel
        - ``NONE``: do not take flags as input

        For backwards compatibility, ``True`` is an alias for ``CHANNEL`` and
        ``False`` is an alias for ``NONE``.
    tuning : mapping, optional
        Kernel tuning parameters; if omitted, will autotune. The possible
        parameters are

        - wgs: number of work-items per baseline
        - csplit: approximate number of workitems for each channel
    """

    host_class = host.BackgroundMedianFilterHost
    autotune_version = 4

    def __init__(self, context: AbstractContext, width: int, is_amplitude: bool = False,
                 use_flags: Union[BackgroundFlags, bool] = BackgroundFlags.NONE,
                 tuning: Optional[Mapping[str, Any]] = None):
        self.context = context
        self.width = width
        self.is_amplitude = is_amplitude
        if use_flags is True:
            use_flags = BackgroundFlags.CHANNEL
        elif use_flags is False:
            use_flags = BackgroundFlags.NONE
        if not isinstance(use_flags, BackgroundFlags):
            raise TypeError('use_flags must be an instance of BackgroundFlags or bool')
        if tuning is None:
            tuning = self.autotune(context, width, is_amplitude, use_flags)
        self.use_flags = use_flags
        self.wgs = tuning['wgs']
        self.csplit = tuning['csplit']
        self.program = accel.build(
            context, 'rfi/background_median_filter.mako',
            {
                'width': width, 'wgs': self.wgs,
                'is_amplitude': is_amplitude,
                'BackgroundFlags': BackgroundFlags,
                'use_flags': use_flags
            })

    @classmethod
    @tune.autotuner(test={'wgs': 128, 'csplit': 4})
    def autotune(cls, context: AbstractContext, width: int, is_amplitude: bool,
                 use_flags: BackgroundFlags) -> Mapping[str, Any]:
        queue = context.create_tuning_command_queue()
        # Note: baselines must be a multiple of any tested workgroup size
        channels = 4096
        baselines = 8192
        shape = (channels, baselines)
        vis_type = np.float32 if is_amplitude else np.complex64
        vis = DeviceArray(context, shape, vis_type)
        deviations = DeviceArray(context, shape, np.float32)
        if use_flags == BackgroundFlags.CHANNEL:
            flags = DeviceArray(context, (channels,), np.uint8)
        elif use_flags == BackgroundFlags.FULL:
            flags = DeviceArray(context, shape, np.uint8)
        # Initialize with Gaussian random values
        rs = np.random.RandomState(seed=1)
        if is_amplitude:
            vis_host: np.ndarray = rs.rayleigh(size=shape).astype(np.float32)
        else:
            vis_host = rs.standard_normal(shape) + rs.standard_normal(shape) * 1j
            vis_host = vis_host.astype(np.complex64)
        vis.set(queue, vis_host)
        if use_flags:
            flags.set(queue, (rs.uniform(size=flags.shape) < 0.0001).astype(np.uint8))

        def generate(**tuning) -> Callable[[int], float]:
            fn = cls(context, width, is_amplitude, use_flags, tuning).instantiate(
                queue, channels, baselines)
            fn.bind(vis=vis, deviations=deviations)
            if use_flags:
                fn.bind(flags=flags)
            return tune.make_measure(queue, fn)

        return tune.autotune(generate, wgs=[32, 64, 128, 256, 512], csplit=[1, 2, 4, 8, 16])

    def instantiate(self, command_queue: AbstractCommandQueue, channels: int, baselines: int,
                    allocator: Optional[AbstractAllocator] = None
                    ) -> 'BackgroundMedianFilterDevice':
        """Create an instance. See :class:`BackgroundMedianFilterDevice`."""
        return BackgroundMedianFilterDevice(self, command_queue, channels, baselines, allocator)


class BackgroundMedianFilterDevice(AbstractBackgroundDevice):
    """Concrete instance of :class:`BackgroundMedianFilterDeviceTemplate`.

    .. rubric:: Slots

    **vis** : channels × baselines, float32 or complex64
        Input visibilities, or their amplitudes if `template.is_amplitude` is true
    **flags** : channels × baselines or channels, float32
        Input flags (only present if `template.use_flags` is used)
    **deviations** : channels × baselines, float32
        Output deviations from the background

    Parameters
    ----------
    template
        Operation template
    command_queue
        Command queue for the operation
    channels, baselines
        Shape of the visibilities array
    allocator
        Allocator used to allocate unbound slots
    """

    def __init__(self, template: BackgroundMedianFilterDeviceTemplate,
                 command_queue: AbstractCommandQueue,
                 channels: int, baselines: int,
                 allocator: Optional[AbstractAllocator] = None) -> None:
        super().__init__(command_queue, allocator)
        self.template = template
        self.kernel = template.program.get_kernel('background_median_filter')
        self.channels = channels
        self.baselines = baselines
        vis_type = np.float32 if template.is_amplitude else np.complex64
        dims = (channels,
                accel.Dimension(baselines, self.template.wgs))
        self.slots['vis'] = accel.IOSlot(dims, vis_type)
        self.slots['deviations'] = accel.IOSlot(dims, np.float32)
        if template.use_flags == BackgroundFlags.FULL:
            self.slots['flags'] = accel.IOSlot(dims, np.uint8)
        elif template.use_flags == BackgroundFlags.CHANNEL:
            self.slots['flags'] = accel.IOSlot((channels,), np.uint8)

    def _run(self) -> None:
        VT = accel.divup(self.channels, self.template.csplit)
        xblocks = accel.divup(self.baselines, self.template.wgs)
        yblocks = accel.divup(self.channels, VT)

        buffers = [self.buffer('vis'), self.buffer('deviations')]
        if self.template.use_flags:
            buffers.append(self.buffer('flags'))
        stride = buffers[0].padded_shape[1]
        self.command_queue.enqueue_kernel(
            self.kernel,
            [b.buffer for b in buffers] + [
                np.int32(self.channels),
                np.int32(stride),
                np.int32(VT)
            ],
            global_size=(xblocks * self.template.wgs, yblocks),
            local_size=(self.template.wgs, 1))

    def parameters(self) -> Mapping[str, Any]:
        return {
            'width': self.template.width,
            'use_flags': self.template.use_flags.name,
            'channels': self.channels,
            'baselines': self.baselines
        }


class NoiseEstHostFromDevice(host.AbstractNoiseEstHost):
    """Wrap a device-side noise estimator template to present the host interface."""

    def __init__(self, template: AbstractNoiseEstDeviceTemplate,
                 command_queue: AbstractCommandQueue) -> None:
        self.template = template
        self.command_queue = command_queue

    def __call__(self, deviations: np.ndarray) -> np.ndarray:
        (channels, baselines) = deviations.shape
        transposed = self.template.transposed
        if transposed:
            deviations = deviations.T
        fn = self.template.instantiate(self.command_queue, channels, baselines)
        # Allocate and populate memory
        fn.ensure_all_bound()
        fn.buffer('deviations').set(self.command_queue, deviations)
        # Perform computation
        fn()
        # Copy back results
        noise = fn.buffer('noise').get(self.command_queue)
        return noise


class NoiseEstMADDeviceTemplate(AbstractNoiseEstDeviceTemplate):
    """Estimate noise using the median of non-zero absolute deviations.

    In most cases :class:`NoiseEstMADTDeviceTemplate` is more efficient.

    Parameters
    ----------
    context
        Context for which kernels will be compiled
    tuning
        Kernel tuning parameters; if omitted, will autotune. The possible
        parameters are

        - wgsx: number of baselines per workgroup
        - wgsy: number of channels per workgroup
    """

    host_class = host.NoiseEstMADHost
    transposed = False

    def __init__(self, context: AbstractContext,
                 tuning: Optional[Mapping[str, Any]] = None) -> None:
        if tuning is None:
            tuning = self.autotune(context)
        self.context = context
        self.wgsx = tuning['wgsx']
        self.wgsy = tuning['wgsy']
        self.program = accel.build(context, 'rfi/madnz.mako',
                                   {'wgsx': self.wgsx, 'wgsy': self.wgsy})

    @classmethod
    @tune.autotuner(test={'wgsx': 32, 'wgsy': 8})
    def autotune(cls, context: AbstractContext) -> Mapping[str, Any]:
        # TODO: do real autotuning
        return {'wgsx': 32, 'wgsy': 8}

    def instantiate(self, command_queue: AbstractCommandQueue,
                    channels: int, baselines: int,
                    allocator: Optional[AbstractAllocator] = None) -> 'NoiseEstMADDevice':
        """Create an instance. See :class:`NoiseEstMADDevice`."""
        return NoiseEstMADDevice(self, command_queue, channels, baselines, allocator)


class NoiseEstMADDevice(AbstractNoiseEstDevice):
    """Concrete instantiation of :class:`NoiseEstMADDeviceTemplate`.

    .. rubric:: Slots

    **deviations** : channels × baselines, float32
        Input deviations from the background, computed by a backgrounder
    **noise** : baselines, float32
        Output per-baseline noise estimate

    Parameters
    ----------
    template
        Operation template
    command_queue
        Command-queue in which work will be enqueued
    channels, baselines
        Shape of the visibility array
    allocator
        Allocator used to allocate unbound slots
    """

    transposed = False

    def __init__(self, template: NoiseEstMADDeviceTemplate,
                 command_queue: AbstractCommandQueue,
                 channels: int, baselines: int,
                 allocator: Optional[AbstractAllocator] = None) -> None:
        super().__init__(command_queue, allocator)
        self.template = template
        self.kernel = template.program.get_kernel('madnz')
        self.channels = channels
        self.baselines = baselines
        baselines_dim = accel.Dimension(baselines, self.template.wgsx)
        self.slots['noise'] = accel.IOSlot((baselines_dim,), np.float32)
        self.slots['deviations'] = accel.IOSlot((channels, baselines_dim), np.float32)

    def _run(self) -> None:
        blocks = accel.divup(self.baselines, self.template.wgsx)
        vt = accel.divup(self.channels, self.template.wgsy)
        deviations = self.buffer('deviations')
        noise = self.buffer('noise')
        self.command_queue.enqueue_kernel(
            self.kernel,
            [
                deviations.buffer, noise.buffer,
                np.int32(self.channels), np.int32(deviations.padded_shape[1]),
                np.int32(vt)
            ],
            global_size=(blocks * self.template.wgsx, self.template.wgsy),
            local_size=(self.template.wgsx, self.template.wgsy))

    def parameters(self) -> Mapping[str, Any]:
        return {
            'channels': self.channels,
            'baselines': self.baselines
        }


class NoiseEstMADTDeviceTemplate(AbstractNoiseEstDeviceTemplate):
    """Device-side noise estimation by median of absolute deviations.

    It should give the same results as :class:`NoiseEstMADHost`, up to
    floating-point accuracy. It uses transposed (baseline-major) memory order,
    which allows an entire baseline to be efficiently loaded into registers.

    .. note:: There is a tradeoff in selecting the workgroup size: a large
        value gives more parallelism and reduces the register pressure, but
        increases the overhead of reduction operations.

    .. note:: This class may fail for very large numbers of channels (10k can
        definitely be supported), in which case :class:`NoiseEstMADDevice` may be
        used.

    Parameters
    ----------
    context
        Context for which kernels will be compiled
    max_channels
        Maximum number of channels. Choosing too large a value will
        reduce performance.
    tuning
        Kernel tuning parameters; if omitted, will autotune. The possible
        parameters are

        - wgsx: number of work-items per baseline
    """

    host_class = host.NoiseEstMADHost
    transposed = True

    def __init__(self, context: AbstractContext, max_channels: int,
                 tuning: Optional[Mapping[str, Any]] = None) -> None:
        self.context = context
        self.max_channels = max_channels
        if tuning is None:
            tuning = self.autotune(context, max_channels)
        self.wgsx = tuning['wgsx']
        vt = accel.divup(max_channels, self.wgsx)
        self.program = accel.build(context, 'rfi/madnz_t.mako', {'vt': vt, 'wgsx': self.wgsx})

    @classmethod
    @tune.autotuner(test={'wgsx': 128})
    def autotune(cls, context: AbstractContext, max_channels: int) -> Mapping[str, Any]:
        queue = context.create_tuning_command_queue()
        baselines = 128
        rs = np.random.RandomState(seed=1)
        host_deviations = rs.uniform(size=(baselines, max_channels)).astype(np.float32)

        def generate(**tuning) -> Callable[[int], float]:
            # Very large values of VT cause the AMD compiler to choke and segfault
            if max_channels > 256 * tuning['wgsx']:
                raise ValueError('wgsx is too small')
            fn = cls(context, max_channels, tuning).instantiate(queue, max_channels, baselines)
            fn.slots['noise'].allocate(fn.allocator)
            deviations = fn.slots['deviations'].allocate(fn.allocator)
            deviations.set(queue, host_deviations)
            return tune.make_measure(queue, fn)

        return tune.autotune(generate, wgsx=[32, 64, 128, 256, 512, 1024])

    def instantiate(self, command_queue: AbstractCommandQueue, channels: int, baselines: int,
                    allocator: Optional[AbstractAllocator] = None) -> 'NoiseEstMADTDevice':
        """Create an instance. See :class:`NoiseEstMADTDevice`."""
        return NoiseEstMADTDevice(self, command_queue, channels, baselines, allocator)


class NoiseEstMADTDevice(AbstractNoiseEstDevice):
    """Concrete instance of :class:`NoiseEstMADTDeviceTemplate`.

    .. rubric:: Slots

    **deviations** : baselines × channels, float32
        Input deviations from the background, computed by a backgrounder
    **noise** : baselines, float32
        Output per-baseline noise estimate

    Parameters
    ----------
    template
        Operation template
    command_queue
        Command-queue in which work will be enqueued
    channels, baselines
        Shape of the visibility array
    allocator
        Allocator used to allocate unbound slots
    """

    transposed = True

    def __init__(self, template: NoiseEstMADTDeviceTemplate,
                 command_queue: AbstractCommandQueue,
                 channels: int, baselines: int,
                 allocator: Optional[AbstractAllocator] = None) -> None:
        super().__init__(command_queue, allocator)
        self.template = template
        self.kernel = template.program.get_kernel('madnz_t')
        if channels > self.template.max_channels:
            raise ValueError('channels exceeds max_channels')
        self.channels = channels
        self.baselines = baselines
        self.slots['noise'] = accel.IOSlot((baselines,), np.float32)
        self.slots['deviations'] = accel.IOSlot((baselines, channels), np.float32)

    def _run(self) -> None:
        deviations = self.buffer('deviations')
        noise = self.buffer('noise')
        self.command_queue.enqueue_kernel(
            self.kernel,
            [
                deviations.buffer, noise.buffer,
                np.int32(self.channels), np.int32(deviations.padded_shape[1])
            ],
            global_size=(self.template.wgsx, self.baselines),
            local_size=(self.template.wgsx, 1))

    def parameters(self) -> Mapping[str, Any]:
        return {
            'max_channels': self.template.max_channels,
            'baselines': self.baselines,
            'channels': self.channels
        }


class ThresholdHostFromDevice(host.AbstractThresholdHost):
    """Wrap a device-side thresholder template to present the host interface."""

    def __init__(self, template: AbstractThresholdDeviceTemplate,
                 command_queue: AbstractCommandQueue, *args, **kwargs) -> None:
        self.template = template
        self.command_queue = command_queue
        self.args = args
        self.kwargs = kwargs

    def __call__(self, deviations: np.ndarray, noise: np.ndarray) -> np.ndarray:
        (channels, baselines) = deviations.shape
        transposed = self.template.transposed
        if transposed:
            deviations = deviations.T

        fn = self.template.instantiate(self.command_queue,
                                       channels, baselines,
                                       *self.args, **self.kwargs)
        # Allocate memory and copy data
        fn.ensure_all_bound()
        fn.buffer('deviations').set(self.command_queue, deviations)
        fn.buffer('noise').set(self.command_queue, noise)
        # Do computation
        fn()
        # Copy back results
        flags = fn.buffer('flags').get(self.command_queue)
        if transposed:
            flags = flags.T
        return flags


class ThresholdSimpleDeviceTemplate(AbstractThresholdDeviceTemplate):
    """Device-side thresholding, operating independently on each sample.

    It should give the same results as :class:`ThresholdSimpleHost`, up to
    floating-point accuracy.

    This class can operate on either transposed or non-transposed inputs,
    depending on a constructor argument.

    Parameters
    ----------
    context
        Context for which kernels will be compiled
    transposed
        Whether inputs and outputs are transposed
    flag_value
        Number stored in returned value to indicate RFI
    tuning
        Kernel tuning parameters; if omitted, will autotune. The possible
        parameters are

        - wgsx: number of baselines (channels if `transposed`) per workgroup
        - wgsy: number of channels (baselines if `transposed`) per workgroup
    """

    host_class = host.ThresholdSimpleHost

    def __init__(self, context: AbstractContext, transposed: bool,
                 flag_value: int = 1, tuning: Optional[Mapping[str, Any]] = None) -> None:
        if tuning is None:
            tuning = self.autotune(context)
        self.context = context
        self.transposed = transposed
        self.wgsx = tuning['wgsx']
        self.wgsy = tuning['wgsy']
        self.flag_value = flag_value
        if transposed:
            source_name = 'rfi/threshold_simple_t.mako'
        else:
            source_name = 'rfi/threshold_simple.mako'
        self.program = accel.build(
            context, source_name,
            {'wgsx': self.wgsx, 'wgsy': self.wgsy,
             'flag_value': flag_value})

    @classmethod
    @tune.autotuner(test={'wgsx': 32, 'wgsy': 4})
    def autotune(cls, context: AbstractContext) -> Mapping[str, Any]:
        # TODO: do real autotuning
        return {'wgsx': 32, 'wgsy': 4}

    def instantiate(self, command_queue: AbstractCommandQueue,
                    channels: int, baselines: int, n_sigma: float,
                    allocator: Optional[AbstractAllocator] = None) -> 'ThresholdSimpleDevice':
        """Create an instance. See :class:`ThresholdSimpleDevice`."""
        return ThresholdSimpleDevice(self, command_queue, channels, baselines, n_sigma, allocator)


class ThresholdSimpleDevice(AbstractThresholdDevice):
    """Concrete instance of :class:`ThresholdSimpleDeviceTemplate`.

    .. rubric:: Slots

    **deviations** : channels × baselines (or transposed), float32
        Input deviations from the background
    **noise** : baselines, float32
        Noise estimates per baseline
    **flags** : channels × baselines (or transposed), uint8
        Output flags

    Parameters
    ----------
    template
        Operation template
    command_queue
        Command-queue in which work will be enqueued
    channels, baselines
        Shape of the visibility array
    n_sigma
        Number of (estimated) standard deviations for the threshold
    allocator
        Allocator used to allocate unbound slots
    """

    def __init__(self, template: ThresholdSimpleDeviceTemplate,
                 command_queue: AbstractCommandQueue,
                 channels: int, baselines: int, n_sigma: float,
                 allocator: Optional[AbstractAllocator] = None) -> None:
        super().__init__(command_queue, allocator)
        self.template = template
        if template.transposed:
            kernel_name = 'threshold_simple_t'
        else:
            kernel_name = 'threshold_simple'
        self.kernel = template.program.get_kernel(kernel_name)
        self.n_sigma = n_sigma
        self.channels = channels
        self.baselines = baselines
        self.transposed = template.transposed

        shape = (baselines, channels) if self.transposed else (channels, baselines)
        dims = (accel.Dimension(shape[0], self.template.wgsy),
                accel.Dimension(shape[1], self.template.wgsx))
        noise_dim = dims[0] if self.transposed else dims[1]

        self.slots['deviations'] = accel.IOSlot(dims, np.float32)
        self.slots['noise'] = accel.IOSlot((noise_dim,), np.float32)
        self.slots['flags'] = accel.IOSlot(dims, np.uint8)

    def _run(self) -> None:
        deviations = self.buffer('deviations')
        noise = self.buffer('noise')
        flags = self.buffer('flags')
        stride = deviations.padded_shape[1]

        global_x = accel.roundup(deviations.shape[1], self.template.wgsx)
        global_y = accel.roundup(deviations.shape[0], self.template.wgsy)
        self.command_queue.enqueue_kernel(
            self.kernel,
            [
                deviations.buffer, noise.buffer, flags.buffer,
                np.int32(stride),
                np.float32(self.n_sigma)
            ],
            global_size=(global_x, global_y),
            local_size=(self.template.wgsx, self.template.wgsy))

    def parameters(self) -> Mapping[str, Any]:
        return {
            'n_sigma': self.n_sigma,
            'flag_value': self.template.flag_value,
            'transposed': self.transposed,
            'channels': self.channels,
            'baselines': self.baselines
        }


class ThresholdSumDeviceTemplate(AbstractThresholdDeviceTemplate):
    """A device version of :class:`katsdpsigproc.rfi.host.ThresholdSumHost`.

    It uses transposed data. Performance will be best with a large work
    group size, because of the stencil-like nature of the computation.

    Parameters
    ----------
    context
        Context for which kernels will be compiled
    n_windows
        Number of window sizes to use
    flag_value
        Number stored in returned value to indicate RFI
    tuning
        Kernel tuning parameters; if omitted, will autotune. The possible
        parameters are

        - wgs: Number of work items to use per work group
        - vt: Number of elements to process in each work item
    """

    host_class = host.ThresholdSumHost
    transposed = True

    def __init__(self, context: AbstractContext, n_windows: int = 4, flag_value: int = 1,
                 tuning: Optional[Mapping[str, Any]] = None) -> None:
        if tuning is None:
            tuning = self.autotune(context, n_windows)
        wgs = tuning['wgs']
        vt = tuning['vt']
        edge_size = 2 ** n_windows - n_windows - 1
        self.chunk = wgs * vt - 2 * edge_size
        assert self.chunk > 0
        self.context = context
        self.n_windows = n_windows
        self.wgs = wgs
        self.vt = vt
        self.flag_value = flag_value
        self.program = accel.build(
            context, 'rfi/threshold_sum.mako',
            {'wgs': self.wgs,
             'vt': self.vt,
             'windows' : self.n_windows,
             'flag_value': self.flag_value})

    @classmethod
    @tune.autotuner(test={'wgs': 128, 'vt': 3})
    def autotune(cls, context: AbstractContext, n_windows: int) -> Mapping[str, Any]:
        queue = context.create_tuning_command_queue()
        channels = 4096
        baselines = 128
        shape = (baselines, channels)
        rs = np.random.RandomState(seed=1)
        deviations = DeviceArray(context, shape, dtype=np.float32)
        deviations.set(queue, rs.uniform(size=deviations.shape).astype(np.float32))
        noise = DeviceArray(context, (baselines,), dtype=np.float32)
        noise.set(queue, rs.uniform(high=0.1, size=noise.shape).astype(np.float32))
        flags = DeviceArray(context, shape, dtype=np.uint8)

        def generate(**tuning) -> Callable[[int], float]:
            template = cls(context, n_windows=n_windows, tuning=tuning)
            fn = template.instantiate(queue, channels, baselines, 11.0)
            fn.bind(deviations=deviations, noise=noise, flags=flags)
            return tune.make_measure(queue, fn)

        return tune.autotune(generate,
                             wgs=[32, 64, 128, 256, 512],
                             vt=[1, 2, 3, 4, 8, 16])

    def instantiate(self, command_queue: AbstractCommandQueue,
                    channels: int, baselines: int, n_sigma: float,
                    threshold_falloff: float = _THRESHOLD_SUM_DEFAULT_THRESHOLD_FALLOFF,
                    allocator: Optional[AbstractAllocator] = None) -> 'ThresholdSumDevice':
        """Create an instance. See :class:`ThresholdSumDevice`."""
        return ThresholdSumDevice(self, command_queue, channels, baselines, n_sigma,
                                  threshold_falloff, allocator)


class ThresholdSumDevice(AbstractThresholdDevice):
    """Concrete instance of :class:`ThresholdSumDeviceTemplate`.

    .. rubric:: Slots

    **deviations** : baselines × channels, float32
        Input deviations from the background
    **noise** : baselines, float32
        Noise estimates per baseline
    **flags** : baselines × channels, uint8
        Output flags

    Parameters
    ----------
    template
        Operation template
    command_queue
        Command-queue in which work will be enqueued
    channels, baselines
        Shape of the visibility array
    n_sigma
        Number of (estimated) standard deviations for the threshold
    threshold_falloff
        Controls rate at which thresholds decrease (ρ in Offringa 2010)
    allocator
        Allocator used to allocate unbound slots
    """

    host_class = host.ThresholdSumHost
    transposed = True
    DEFAULT_THRESHOLD_FALLOFF = _THRESHOLD_SUM_DEFAULT_THRESHOLD_FALLOFF

    def __init__(self, template: ThresholdSumDeviceTemplate,
                 command_queue: AbstractCommandQueue,
                 channels: int, baselines: int, n_sigma: float,
                 threshold_falloff: float = DEFAULT_THRESHOLD_FALLOFF,
                 allocator: Optional[AbstractAllocator] = None) -> None:
        super().__init__(command_queue, allocator)
        self.template = template
        self.kernel = template.program.get_kernel('threshold_sum')
        self.channels = channels
        self.baselines = baselines
        self.n_sigma = [np.float32(n_sigma * pow(threshold_falloff, -i))
                        for i in range(template.n_windows)]
        # For channels we must construct the Dimension here rather than in
        # the IOSlot constructor, so that the deviations and flags share
        # the same object and hence have the same stride.
        dims = (baselines, accel.Dimension(channels))
        self.slots['deviations'] = accel.IOSlot(dims, np.float32)
        self.slots['noise'] = accel.IOSlot((baselines,), np.float32)
        self.slots['flags'] = accel.IOSlot(dims, np.uint8)

    def _run(self) -> None:
        deviations = self.buffer('deviations')
        noise = self.buffer('noise')
        flags = self.buffer('flags')

        blocks = accel.divup(self.channels, self.template.chunk)
        args = [deviations.buffer, noise.buffer, flags.buffer,
                np.int32(self.channels), np.int32(deviations.padded_shape[1])]
        args.extend(self.n_sigma)
        self.command_queue.enqueue_kernel(
            self.kernel, args,
            global_size=(blocks * self.template.wgs, self.baselines),
            local_size=(self.template.wgs, 1))

    def parameters(self) -> Mapping[str, Any]:
        return {
            'n_sigma': self.n_sigma,
            'flag_value': self.template.flag_value,
            'channels': self.channels,
            'baselines': self.baselines
        }


class FlaggerDeviceTemplate:
    """Combine device backgrounder, noise estimation and thresholder to create a flagger.

    The thresholder and/or noise estimation may take transposed input, in which
    case this object will manage temporary buffers and the transposition
    automatically.

    Parameters
    ----------
    background, noise_est, threshold
        The templates for the individual steps
    """

    def __init__(self,
                 background: AbstractBackgroundDeviceTemplate,
                 noise_est: AbstractNoiseEstDeviceTemplate,
                 threshold: AbstractThresholdDeviceTemplate) -> None:
        self.background = background
        self.noise_est = noise_est
        self.threshold = threshold
        context = self.background.context
        # Check that all operations work in the same context
        assert self.noise_est.context is context
        assert self.threshold.context is context

        # Create transposition operations if needed
        if noise_est.transposed or threshold.transposed:
            self.transpose_deviations = \
                transpose.TransposeTemplate(
                    context, np.float32, 'float')   # type: Optional[transpose.TransposeTemplate]
        else:
            self.transpose_deviations = None
        if threshold.transposed:
            self.transpose_flags = \
                transpose.TransposeTemplate(
                    context, np.uint8,
                    'unsigned char')      # type: Optional[transpose.TransposeTemplate]
        else:
            self.transpose_flags = None

    def instantiate(self, command_queue: AbstractCommandQueue,
                    channels: int, baselines: int,
                    background_args: Mapping[str, Any] = {},
                    noise_est_args: Mapping[str, Any] = {},
                    threshold_args: Mapping[str, Any] = {},
                    allocator: Optional[AbstractAllocator] = None) -> 'FlaggerDevice':
        """Create an instance (see :class:`FlaggerDevice`)."""
        return FlaggerDevice(self, command_queue, channels, baselines,
                             background_args, noise_est_args, threshold_args,
                             allocator)


class FlaggerDevice(accel.OperationSequence):
    """Concrete instance of :class:`FlaggerDeviceTemplate`.

    .. rubric:: Slots

    **vis** : channels × baselines, float32 or complex64
        Input visibilities (or amplitudes, if the backgrounder takes amplitudes)
    **noise** : baselines, float32
        Estimate of per-baseline noise
    **flags** : channels × baselines, uint8
        Output flags
    **input_flags** : channels × baseline or channels, uint8
        Input flags. These are only used for estimating the
        background, and are *not* automatically copied into the output. Any
        visibilities marked as flagged here will *not* be flagged as RFI.

        This slot is only present if the backgrounder template has
        `use_flags` set, and the value determines the shape.

    .. rubric:: Temporary slots

    Temporary buffers are presented as slots, which allows them to either
    be set by the user or allocated automatically on first use.

    **deviations** : channels × baselines, float32
        Deviations from the background
    **deviations_t** : baselines × channels, float32, optional
        Transpose of `deviations`
    **flags_t** : baselines × channels, uint8, optional
        Transpose of `flags`

    Parameters
    ----------
    template
        Operation template
    command_queue
        Command queue for the operation
    channels, baselines
        Shape of the visibilities array
    background_args
        Extra keyword arguments to pass to the background instantiation
    noise_est_args
        Extra keyword arguments to pass to the noise estimation instantiation
    threshold_args
        Extra keyword arguments to pass to the threshold instantiation
    allocator
        Allocator used to allocate unbound slots
    """

    def __init__(self, template: FlaggerDeviceTemplate,
                 command_queue: AbstractCommandQueue,
                 channels: int, baselines: int,
                 background_args: Mapping[str, Any] = {},
                 noise_est_args: Mapping[str, Any] = {},
                 threshold_args: Mapping[str, Any] = {},
                 allocator: Optional[AbstractAllocator] = None) -> None:
        self.template = template
        self.channels = channels
        self.baselines = baselines
        self.background = self.template.background.instantiate(    # type: ignore
            command_queue, channels, baselines, allocator=allocator, **background_args)
        self.noise_est = self.template.noise_est.instantiate(      # type: ignore
            command_queue, channels, baselines, allocator=allocator, **noise_est_args)
        self.threshold = self.template.threshold.instantiate(      # type: ignore
            command_queue, channels, baselines, allocator=allocator, **threshold_args)

        noise_est_suffix = '_t' if self.noise_est.transposed else ''
        threshold_suffix = '_t' if self.threshold.transposed else ''

        operations = []     # type: List[Tuple[str, accel.Operation]]
        compounds = {
            'vis': ['background:vis'],
            'input_flags': ['background:flags'],
            'deviations': ['background:deviations', 'transpose_deviations:src'],
            'deviations_t': ['transpose_deviations:dest'],
            'noise': ['noise_est:noise', 'threshold:noise'],
            'flags_t': ['transpose_flags:src'],
            'flags': ['transpose_flags:dest'],
        }
        compounds['deviations' + noise_est_suffix].append('noise_est:deviations')
        compounds['deviations' + threshold_suffix].append('threshold:deviations')
        compounds['flags' + threshold_suffix].append('threshold:flags')

        operations.append(('background', self.background))
        if self.template.transpose_deviations:
            self.transpose_deviations = self.template.transpose_deviations.instantiate(
                command_queue, (channels, baselines))
            operations.append(('transpose_deviations', self.transpose_deviations))
        operations.append(('noise_est', self.noise_est))
        operations.append(('threshold', self.threshold))
        if self.template.transpose_flags:
            self.transpose_flags = self.template.transpose_flags.instantiate(
                command_queue, (baselines, channels))
            operations.append(('transpose_flags', self.transpose_flags))

        super().__init__(
            command_queue, operations, compounds, allocator=allocator)


class FlaggerHostFromDevice(host.AbstractFlaggerHost):
    """Wrapper that makes a :class:`FlaggerDeviceTemplate` look like a :class:`FlaggerHost`.

    This is intended only for ease of use. It is not efficient, because it
    allocates and frees memory on every call.

    Parameters
    ----------
    template
        Operation template
    command_queue
        Command queue for the operation
    background_args
        Extra keyword arguments to pass to the background instantiation
    noise_est_args
        Extra keyword arguments to pass to the noise estimation instantiation
    threshold_args
        Extra keyword arguments to pass to the threshold instantiation
    """

    def __init__(self, template: FlaggerDeviceTemplate,
                 command_queue: AbstractCommandQueue,
                 background_args: Mapping[str, Any] = {},
                 noise_est_args: Mapping[str, Any] = {},
                 threshold_args: Mapping[str, Any] = {}) -> None:
        self.template = template
        self.command_queue = command_queue
        self.background_args = dict(background_args)
        self.noise_est_args = dict(noise_est_args)
        self.threshold_args = dict(threshold_args)

    def __call__(self, vis: np.ndarray,
                 input_flags: Optional[np.ndarray] = None) -> np.ndarray:
        if input_flags is not None and not self.template.background.use_flags:
            raise TypeError("channel flags were provided but not included in the template")
        if input_flags is None and self.template.background.use_flags:
            raise TypeError("channel flags were expected but not provided")
        (channels, baselines) = vis.shape
        fn = self.template.instantiate(
            self.command_queue, channels, baselines,
            self.background_args, self.noise_est_args, self.threshold_args)
        fn.ensure_all_bound()
        fn.buffer('vis').set(self.command_queue, vis)
        if input_flags is not None:
            fn.buffer('input_flags').set(self.command_queue, input_flags)
        fn()
        return fn.buffer('flags').get(self.command_queue)
